"""coBib's commands.

coBib provides a series of commands, all of which are available both, from the command-line as well
as through the text user interface (see also `cobib.ui.tui`).
The abstract interface which should be implemented is defined by the `cobib.commands.base_command`.
"""

from .add import AddCommand as AddCommand
from .delete import DeleteCommand as DeleteCommand
from .edit import EditCommand as EditCommand
from .export import ExportCommand as ExportCommand
from .import_ import ImportCommand as ImportCommand
from .init import InitCommand as InitCommand
from .list_ import ListCommand as ListCommand
from .modify import ModifyCommand as ModifyCommand
from .open import OpenCommand as OpenCommand
from .redo import RedoCommand as RedoCommand
from .search import SearchCommand as SearchCommand
from .show import ShowCommand as ShowCommand
from .undo import UndoCommand as UndoCommand



"""Tests for coBib's ListCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import contextlib
import os
from copy import copy
from io import StringIO
from itertools import zip_longest
from shutil import copyfile
from typing import TYPE_CHECKING, Any, List, Set, Type

import pytest
from rich.table import Table
from typing_extensions import override

from cobib.commands import ListCommand
from cobib.config import Event, config
from cobib.database import Database

from .. import get_resource
from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestListCommand(CommandTest):
    """Tests for coBib's ListCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return ListCommand

    @pytest.mark.parametrize(
        ["args", "expected_labels", "config_overwrite"],
        [
            [[], ["einstein", "latexcompanion", "knuthwebsite"], False],
            [["-r"], ["knuthwebsite", "latexcompanion", "einstein"], False],
            [["-s", "year"], ["knuthwebsite", "einstein", "latexcompanion"], False],
            [["-r", "-s", "year"], ["latexcompanion", "einstein", "knuthwebsite"], False],
            [["++author", "Einstein"], ["einstein"], False],
            [["++author", "einstein", "-i"], ["einstein"], False],
            [["++author", "einstein", "-I"], [], True],
            [["--author", "Einstein"], ["latexcompanion", "knuthwebsite"], False],
            [["++author", "Einstein", "++author", "Knuth"], [], False],
            [
                ["-x", "++author", "Einstein", "++author", "Knuth"],
                ["einstein", "knuthwebsite"],
                False,
            ],
        ],
    )
    def test_command(
        self, setup: Any, args: List[str], expected_labels: List[str], config_overwrite: bool
    ) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            args: the arguments to pass to the command.
            expected_labels: the expected list of labels.
            config_overwrite: what to overwrite `config.commands.list_.ignore_case` with.
        """
        config.commands.list_.ignore_case = config_overwrite
        original_default_columns = copy(config.commands.list_.default_columns)

        cmd = ListCommand(*args)
        cmd.execute()
        assert [entry.label for entry in cmd.entries] == expected_labels
        # NOTE: this is a regression test against https://gitlab.com/cobib/cobib/-/issues/117
        assert original_default_columns == config.commands.list_.default_columns

    @pytest.mark.parametrize(
        ["args", "expected_labels", "expected_keys", "config_overwrite"],
        [
            [[], ["einstein", "latexcompanion", "knuthwebsite"], set(), False],
            [["++author", "Einstein"], ["einstein"], {"author"}, False],
            [["++author", "einstein", "-i"], ["einstein"], {"author"}, False],
            [["++author", "einstein", "-I"], [], {"author"}, True],
            [["--author", "Einstein"], ["latexcompanion", "knuthwebsite"], {"author"}, False],
            [["++author", "Einstein", "++author", "Knuth"], [], {"author"}, False],
            [
                ["-x", "++author", "Einstein", "++author", "Knuth"],
                ["einstein", "knuthwebsite"],
                {"author"},
                False,
            ],
        ],
    )
    def test_filter_entries(
        self,
        setup: Any,
        args: List[str],
        expected_labels: List[str],
        expected_keys: Set[str],
        config_overwrite: bool,
    ) -> None:
        """Tests the filtering methods.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            args: the arguments to pass to the command.
            expected_labels: the expected list of labels which match the filter.
            expected_keys: the expected keys which were filtered on.
            config_overwrite: what to overwrite `config.commands.list_.ignore_case` with.
        """
        config.commands.list_.ignore_case = config_overwrite

        filtered_entries, filtered_keys = ListCommand(*args).filter_entries()
        assert filtered_keys == expected_keys
        assert [entry.label for entry in filtered_entries] == expected_labels

    def test_missing_keys(self, setup: Any) -> None:
        """Asserts issue #1 is fixed.

        When a key is queried which is not present in all entries, the list command should return
        normally.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        filtered_entries, filtered_keys = ListCommand("++year", "1905").filter_entries()
        assert [entry.label for entry in filtered_entries] == ["einstein"]
        assert filtered_keys == {"year"}

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "expected",
        [
            [
                ["label", "title"],
                ["einstein", r"Zur Elektrodynamik bewegter K{\"o}rper"],
                ["latexcompanion", r"The \LaTeX\ Companion"],
                ["knuthwebsite", "Knuth: Computers and Typesetting"],
            ],
        ],
    )
    # other variants are already covered by test_command
    async def test_cmdline(
        self,
        setup: Any,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
        expected: List[List[str]],
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
            expected: the expected list of labels.
        """
        await self.run_module(monkeypatch, "main", ["cobib", "--porcelain", "list"])
        stdout = capsys.readouterr().out.strip().split("\n")
        for line, truth in zip_longest(stdout, expected):
            assert line.split("::") == truth

    @pytest.mark.parametrize(
        ["args", "expected_cols", "expected_rows"],
        [
            [[], ["label", "title"], ["einstein", "latexcompanion", "knuthwebsite"]],
            [["-r"], ["label", "title"], ["knuthwebsite", "latexcompanion", "einstein"]],
            [
                ["-s", "year"],
                ["label", "title", "year"],
                ["knuthwebsite", "einstein", "latexcompanion"],
            ],
            [
                ["-r", "-s", "year"],
                ["label", "title", "year"],
                ["latexcompanion", "einstein", "knuthwebsite"],
            ],
            [["++author", "Einstein"], ["label", "title", "author"], ["einstein"]],
            [
                ["--author", "Einstein"],
                ["label", "title", "author"],
                ["latexcompanion", "knuthwebsite"],
            ],
            [["++author", "Einstein", "++author", "Knuth"], ["label", "title", "author"], []],
            [
                ["-x", "++author", "Einstein", "++author", "Knuth"],
                ["label", "title", "author"],
                ["einstein", "knuthwebsite"],
            ],
        ],
    )
    def test_render_rich(
        self, setup: Any, args: List[str], expected_cols: List[str], expected_rows: List[str]
    ) -> None:
        """Test the `render_rich` method.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            args: the arguments to pass to the command.
            expected_cols: the expected list of columns.
            expected_rows: the expected list of row entries.
        """
        cmd = ListCommand(*args)
        cmd.execute()
        renderable = cmd.render_rich()

        assert isinstance(renderable, Table)
        assert [col.header for col in renderable.columns] == expected_cols
        assert len(renderable.rows) == len(expected_rows)
        # pylint: disable=protected-access
        assert renderable.columns[0]._cells == expected_rows

    # manually overwrite this test because we must populate the database with actual data
    def test_handle_argument_error(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test handling of ArgumentError.

        Args:
            caplog: the built-in pytest fixture.
        """
        # use temporary config
        config.database.file = self.COBIB_TEST_DIR / "database.yaml"
        config.database.git = True

        # load temporary database
        os.makedirs(self.COBIB_TEST_DIR, exist_ok=True)
        copyfile(get_resource("example_literature.yaml"), config.database.file)
        Database().read()

        try:
            super().test_handle_argument_error(caplog)
        except SystemExit:
            pass
        finally:
            # clean up file system
            os.remove(config.database.file)
            # clean up database
            Database().clear()
            # clean up config
            config.defaults()

    def test_event_pre_list_command(self, setup: Any) -> None:
        """Tests the PreListCommand event."""

        @Event.PreListCommand.subscribe
        def hook(command: ListCommand) -> None:
            command.largs.author = None

        assert Event.PreListCommand.validate()

        cmd = ListCommand("++author", "Einstein")
        cmd.execute()

        assert [entry.label for entry in cmd.entries] == [
            "einstein",
            "latexcompanion",
            "knuthwebsite",
        ]

    def test_event_post_list_command(self, setup: Any) -> None:
        """Tests the PostListCommand event."""

        @Event.PostListCommand.subscribe
        def hook(command: ListCommand) -> None:
            print([entry.label for entry in command.entries], command.columns)

        assert Event.PostListCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as out:
            cmd = ListCommand()
            cmd.execute()

            assert (
                out.getvalue()
                == "['einstein', 'latexcompanion', 'knuthwebsite'] ['label', 'title']\n"
            )



"""coBib's command-line test class."""

from __future__ import annotations

import os
import runpy
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    import pytest


class CmdLineTest:
    """A command-line test runs coBib's command-line interface."""

    @staticmethod
    async def run_module(monkeypatch: pytest.MonkeyPatch, main: str, sys_argv: List[str]) -> None:
        """Gets the coBib runtime module after monkeypatching `sys.argv`.

        Args:
            monkeypatch: the built-in pytest fixture.
            main: the name of the `main` executable of the module to run.
            sys_argv: the list of values with which to monkeypatch `sys.argv`.
        """
        os.environ["COBIB_CONFIG"] = "0"
        monkeypatch.setattr("sys.argv", sys_argv)
        module = runpy.run_module("cobib")
        await module.get(main)()  # type: ignore



"""Tests for coBib's Entry class."""

from typing import Any, Dict, List, Tuple

import pytest

from cobib.database import Entry
from cobib.parsers.bibtex import BibtexParser
from cobib.utils.rel_path import RelPath

from .. import get_resource

EXAMPLE_BIBTEX_FILE = get_resource("example_entry.bib")
EXAMPLE_YAML_FILE = get_resource("example_entry.yaml")

EXAMPLE_ENTRY_DICT = {
    "ENTRYTYPE": "article",
    "author": "Yudong Cao and Jonathan Romero and Jonathan P. Olson and Matthias Degroote and Peter"
    + " D. Johnson and M{\\'a}ria Kieferov{\\'a} and Ian D. Kivlichan and Tim Menke and Borja "
    + "Peropadre and Nicolas P. D. Sawaya and Sukin Sim and Libor Veis and Al{\\'a}n Aspuru-Guzik",
    "doi": "10.1021/acs.chemrev.8b00803",
    "journal": "Chemical Reviews",
    "month": "aug",
    "number": 19,
    "pages": "10856--10915",
    "publisher": "American Chemical Society ({ACS})",
    "title": "Quantum Chemistry in the Age of Quantum Computing",
    "url": "https://doi.org/10.1021%2Facs.chemrev.8b00803",
    "volume": 119,
    "year": 2019,
}


def test_init_logging(caplog: pytest.LogCaptureFixture) -> None:
    """Test init logging for linting purposes.

    Args:
        caplog: the built-in pytest fixture.
    """
    entry = Entry("dummy", {"ID": "dummy", "number": "1"})
    assert entry.data["number"] == 1
    assert (
        "cobib.database.entry",
        20,
        "Converting field 'number' of entry 'dummy' to integer: 1.",
    ) in caplog.record_tuples
    assert entry.label == "dummy"
    assert "ID" not in entry.data
    assert (
        "cobib.database.entry",
        20,
        "The field 'ID' of entry 'dummy' is no longer required. It will be inferred from the entry "
        "label.",
    ) in caplog.record_tuples


def test_equality() -> None:
    """Test entry equality."""
    entry_1 = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    entry_2 = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    # assert mutability of Entry
    assert entry_1 is not entry_2
    # assert equality of entries
    assert entry_1 == entry_2


def test_entry_set_label() -> None:
    """Test label changing."""
    # this test may fail if the input dict is not copied
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    entry.label = "Cao2019"
    assert entry.label == "Cao2019"


def test_entry_set_tags(caplog: pytest.LogCaptureFixture) -> None:
    """Test tags setting.

    Args:
        caplog: the built-in pytest fixture.
    """
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    assert entry.tags == []
    # NB: tags must be a list
    entry.tags = ["foo"]
    assert entry.tags == ["foo"]
    # list works as expected
    entry.tags = ["foo", "bar"]
    assert entry.tags == ["foo", "bar"]
    # check lint logging
    entry.tags = "foo, bar"  # type: ignore
    assert entry.tags == ["foo", "bar"]
    assert (
        "cobib.database.entry",
        20,
        "Converted the field 'tags' of entry 'Cao_2019' to a list. You can consider storing it as "
        "such directly.",
    ) in caplog.record_tuples


@pytest.mark.parametrize(
    "files",
    [
        [EXAMPLE_BIBTEX_FILE],
        [EXAMPLE_BIBTEX_FILE, EXAMPLE_YAML_FILE],
    ],
)
def test_entry_set_file(files: List[str], caplog: pytest.LogCaptureFixture) -> None:
    """Test file setting.

    Args:
        files: a list of paths to files.
        caplog: the built-in pytest fixture.
    """
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    entry.file = files[0] if len(files) == 1 else files  # type: ignore
    expected = [str(RelPath(file)) for file in files]
    assert entry.file == expected
    # check lint logging
    if len(files) > 1:
        entry.file = ", ".join(files)  # type: ignore
        assert entry.file == expected
        assert (
            "cobib.database.entry",
            20,
            "Converted the field 'file' of entry 'Cao_2019' to a list. You can consider storing it "
            "as such directly.",
        ) in caplog.record_tuples


def test_entry_set_url(caplog: pytest.LogCaptureFixture) -> None:
    """Test url setting.

    Args:
        caplog: the built-in pytest fixture.
    """
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    entry.url = "https://dummy.org/, https://dummy.com/"  # type: ignore
    assert entry.url == ["https://dummy.org/", "https://dummy.com/"]
    assert (
        "cobib.database.entry",
        20,
        "Converted the field 'url' of entry 'Cao_2019' to a list. You can consider storing it as "
        "such directly.",
    ) in caplog.record_tuples


@pytest.mark.parametrize(
    ["month", "expected"],
    [
        [(1, "January"), "jan"],
        [(2, "February"), "feb"],
        [(3, "March"), "mar"],
        [(4, "April"), "apr"],
        [(5, "May"), "may"],
        [(6, "June"), "jun"],
        [(7, "July"), "jul"],
        [(8, "August"), "aug"],
        [(9, "September"), "sep"],
        [(10, "October"), "oct"],
        [(11, "November"), "nov"],
        [(12, "December"), "dec"],
    ],
)
def test_entry_set_month(
    month: Tuple[int, str], expected: str, caplog: pytest.LogCaptureFixture
) -> None:
    """Test month setting.

    Args:
        month: a pair containing the month index and full name.
        expected: the expected three-letter code of the month.
        caplog: the built-in pytest fixture.
    """
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    assert entry.data["month"] == "aug"
    entry.month = month[0]  # type: ignore
    assert entry.data["month"] == expected
    assert (
        "cobib.database.entry",
        20,
        f"Converting field 'month' of entry 'Cao_2019' from '{month[0]}' to '{expected}'.",
    ) in caplog.record_tuples
    entry.month = month[1]
    assert entry.data["month"] == expected
    assert (
        "cobib.database.entry",
        20,
        f"Converting field 'month' of entry 'Cao_2019' from '{month[1]}' to '{expected}'.",
    ) in caplog.record_tuples


@pytest.mark.parametrize(
    ["filter_", "or_", "ignore_case"],
    [
        [{("author", True): ["cao"]}, False, True],
        [{("author", True): ["Cao"]}, False, False],
        [{("author", False): ["wrong_author"]}, False, False],
        [{("author", True): ["cao"], ("year", True): ["2019"]}, False, True],
        [{("author", True): ["Cao"], ("year", True): ["2019"]}, False, False],
        [{("author", True): ["Cao"], ("year", True): ["2020"]}, True, False],
        [{("author", True): ["wrong_author"], ("year", True): ["2019"]}, True, False],
        [{("author", False): ["wrong_author"], ("year", True): ["2019"]}, False, False],
        [{("label", True): [r"\D+_\d+"]}, True, False],
    ],
)
def test_entry_matches(filter_: Dict[Tuple[str, bool], Any], or_: bool, ignore_case: bool) -> None:
    """Test match filter.

    Args:
        filter_: a filter as explained be `cobib.database.Entry.matches`.
        or_: whether to use logical `OR` rather than `AND` for filter combination.
        ignore_case: whether to perform the filter matching case *in*sensitive.
    """
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    # author must match
    assert entry.matches(filter_, or_=or_, ignore_case=ignore_case)


def test_match_with_wrong_key() -> None:
    """Asserts issue #1 is fixed.

    When `cobib.database.Entry.matches` is called with a key in the filter which does not exist in
    the entry, the key should be ignored and the function should return normally.
    """
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    filter_ = {("tags", False): ["test"]}
    assert entry.matches(filter_, or_=False)


@pytest.mark.parametrize(
    ["query", "context", "ignore_case", "expected"],
    [
        [
            ["search_query"],
            1,
            False,
            [
                ["@article{search_dummy,", " abstract = {search_query", "something else"],
                ["something else", "search_query", "something else"],
            ],
        ],
        [
            ["search_query"],
            1,
            True,
            [
                ["@article{search_dummy,", " abstract = {search_query", "something else"],
                ["something else", "Search_Query", "something else"],
                ["something else", "search_query", "something else"],
                ["something else", "Search_Query", "something else}"],
            ],
        ],
        [
            ["[sS]earch_[qQ]uery"],
            1,
            False,
            [
                ["@article{search_dummy,", " abstract = {search_query", "something else"],
                ["something else", "Search_Query", "something else"],
                ["something else", "search_query", "something else"],
                ["something else", "Search_Query", "something else}"],
            ],
        ],
        [
            ["search_query"],
            2,
            False,
            [
                [
                    "@article{search_dummy,",
                    " abstract = {search_query",
                    "something else",
                    "Search_Query",
                ],
                [
                    "Search_Query",
                    "something else",
                    "search_query",
                    "something else",
                    "Search_Query",
                ],
            ],
        ],
        # the following will look almost identical to the second scenarios because otherwise the
        # next match would be included within the context.
        [
            ["search_query"],
            2,
            True,
            [
                ["@article{search_dummy,", " abstract = {search_query", "something else"],
                ["something else", "Search_Query", "something else"],
                ["something else", "search_query", "something else"],
                ["something else", "Search_Query", "something else}", "}"],
            ],
        ],
        # what we care about here, is that the second match does not include lines which occur
        # *before* the first match.
        [
            ["search_query"],
            10,
            False,
            [
                [
                    "@article{search_dummy,",
                    " abstract = {search_query",
                    "something else",
                    "Search_Query",
                    "something else",
                ],
                [
                    "something else",
                    "Search_Query",
                    "something else",
                    "search_query",
                    "something else",
                    "Search_Query",
                    "something else}",
                    "}",
                    "",
                ],
            ],
        ],
        [
            ["query", "Query"],
            1,
            False,
            [
                ["@article{search_dummy,", " abstract = {search_query", "something else"],
                ["something else", "search_query", "something else"],
                ["something else", "Search_Query", "something else"],
                ["something else", "Search_Query", "something else}"],
            ],
        ],
    ],
)
def test_search(
    query: List[str], context: int, ignore_case: bool, expected: List[List[str]]
) -> None:
    """Test search method.

    Args:
        query: the string to search for.
        context: the number of lines to provide as context for the search results.
        ignore_case: whether to perform a case-insensitive search.
        expected: the expected lines.
    """
    entry = Entry(
        "search_dummy",
        {
            "ENTRYTYPE": "article",
            "abstract": "\n".join(
                [
                    "search_query",
                    "something else",
                    "Search_Query",
                    "something else",
                ]
                * 2
            ),
        },
    )
    results = entry.search(query, context=context, ignore_case=ignore_case)
    assert results == expected


def test_search_with_file() -> None:
    """Test the `cobib.database.Entry.search` method with associated file."""
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    entry.file = EXAMPLE_YAML_FILE  # type: ignore
    results = entry.search(["Chemical"], context=0)
    expected = [
        [" journal = {Chemical Reviews},"],
        [" publisher = {American Chemical Society ({ACS})},"],
        ["journal: Chemical Reviews"],
        ["publisher: American Chemical Society ({ACS})"],
    ]
    for res, exp in zip(results, expected):
        assert res == exp


def test_search_with_missing_file(caplog: pytest.LogCaptureFixture) -> None:
    """Test the `cobib.database.Entry.search` method with a missing file.

    Args:
        caplog: the built-in pytest fixture.
    """
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    entry.file = "some_non_existent_file.txt"  # type: ignore
    _ = entry.search(["Chemical"], context=0)
    for source, level, message in caplog.record_tuples:
        if level != 30 or source != "cobib.database.entry":
            continue
        if message.startswith("The associated file") and message.endswith(
            "of entry Cao_2019 does not exist!"
        ):
            break
    else:
        pytest.fail("Missing file was not logged.")


def test_escape_special_chars() -> None:
    """Test escaping of special characters.

    This also ensures that special characters in the label remain unchanged.
    """
    reference = {
        "ENTRYTYPE": "book",
        "title": 'LaTeX Einf{\\"u}hrung',
    }
    entries = BibtexParser().parse(get_resource("example_entry_umlaut.bib", "database"))
    entry = list(entries.values())[0]
    entry.escape_special_chars()
    assert entry.data == reference
    assert entry.label == "LaTeX_Einführung"


def test_save() -> None:
    """Test the `cobib.database.Entry.save` method."""
    entry = Entry("Cao_2019", EXAMPLE_ENTRY_DICT)
    entry_str = entry.save()
    with open(EXAMPLE_YAML_FILE, "r", encoding="utf-8") as expected:
        for line, truth in zip(entry_str.split("\n"), expected):
            assert line == truth.strip("\n")


def test_stringify() -> None:
    """Test the `cobib.database.Entry.stringify` method."""
    entry = Entry(
        "dummy",
        {
            "file": ["/tmp/a.txt", "/tmp/b.txt"],
            "month": 8,
            "tags": ["tag1", "tag2"],
        },
    )
    expected = {
        "label": "dummy",
        "file": "/tmp/a.txt, /tmp/b.txt",
        "month": "aug",
        "tags": "tag1, tag2",
    }
    assert entry.stringify() == expected



"""coBib's URL parser.

This parser is capable of generating `cobib.database.Entry` instances from general URLs.
This parser checks the URL for contained arXiv IDs, DOIs, and ISBNs in this order.
If none of the above match (or fail to return an `Entry`) it will fall back and extract all DOIs
from the website which the URL is pointing to. It will then use the most common DOI (if it occurs
more often than once) and use that as the DOI to which this URL supposedly redirects.

The parser is registered under the `--url` command-line argument of the
`cobib.commands.add.AddCommand`.

The following documentation is mostly inherited from the abstract interface
`cobib.parsers.base_parser`.
"""

import logging
import re
from collections import Counter, OrderedDict
from typing import Dict

import requests
from typing_extensions import override

from cobib.config import Event
from cobib.database import Entry

from .arxiv import ARXIV_REGEX, ArxivParser
from .base_parser import Parser
from .doi import DOI_REGEX, DOIParser
from .isbn import ISBN_REGEX, ISBNParser

LOGGER = logging.getLogger(__name__)


class URLParser(Parser):
    # pylint: disable=too-many-return-statements
    """The URL Parser."""

    name = "url"

    @override
    def parse(self, string: str) -> Dict[str, Entry]:
        string = Event.PreURLParse.fire(string) or string

        if re.search(ARXIV_REGEX, string):
            LOGGER.debug("URL contains an arXiv ID")
            entries = ArxivParser().parse(string)
            if entries:
                LOGGER.debug("Successfully extracted metadata from URL with ArxivParser")
                return entries
        if re.search(DOI_REGEX, string):
            LOGGER.debug("URL contains a DOI")
            entries = DOIParser().parse(string)
            if entries:
                LOGGER.debug("Successfully extracted metadata from URL with DOIParser")
                return entries
        if re.search(ISBN_REGEX, string):
            LOGGER.debug("URL contains an ISBN")
            entries = ISBNParser().parse(string)
            if entries:
                LOGGER.debug("Successfully extracted metadata from URL with ISBNParser")
                return entries

        try:
            page = requests.get(string, timeout=10)
        except requests.exceptions.RequestException as err:
            LOGGER.error("An Exception occurred while trying to query the URL: %s.", string)
            LOGGER.error(err)
            return OrderedDict()

        LOGGER.debug("Falling back to determining most common DOI in URLs page contents")
        matches = re.findall(DOI_REGEX, page.text)
        dois = Counter(matches)
        if not dois:
            LOGGER.error("Could not find any DOIs on the URLs page: %s", string)
            return OrderedDict()
        # we assume the most common DOI on the page is the one which we are looking for
        most_common_doi = dois.most_common(1)[0]
        LOGGER.debug("Most common DOI is: %s", most_common_doi)
        if most_common_doi[1] > 1:
            entries = DOIParser().parse(most_common_doi[0])

        if entries:
            Event.PostURLParse.fire(entries)

            LOGGER.debug("Successfully extracted metadata from most common DOI")
            return entries

        LOGGER.error("Could not extract metadata from URL: %s", string)
        return OrderedDict()

    def dump(self, entry: Entry) -> None:
        """We cannot dump a generic entry as a URL."""
        LOGGER.error("Cannot dump an entry as a URL.")



"""coBib's Database class."""

from __future__ import annotations

import logging
import re
import sys
from collections import OrderedDict
from typing import TYPE_CHECKING, Dict, List, Optional, cast

from cobib.config import config
from cobib.utils.rel_path import RelPath

if TYPE_CHECKING:
    import cobib.database

LOGGER = logging.getLogger(__name__)


# TODO: once Python 3.9 becomes the default, OrderedDict can be properly sub-typed
class Database(OrderedDict):  # type: ignore
    """coBib's Database class is a runtime interface to the plain-test YAML file.

    This class is a *singleton*.
    Thus, accessing `cobib.database.Database` will always yield the identical instance at runtime.
    This ensures data consistency during all operations on the database.
    """

    _instance: Optional[Database] = None
    """The singleton instance of this class."""

    _unsaved_entries: Dict[str, Optional[str]] = {}
    """A dictionary of changed entries which have not been written to the database file, yet.
    The keys are the entry labels. If it the entry was removed, the value of this key is `None`.
    Otherwise it is set to the label of the changed entry (which may be different from the previous
    label, indicating a renaming of the entry)."""

    def __new__(cls) -> Database:
        """Singleton constructor.

        This method gets called when accessing `Database` and enforces the singleton pattern
        implemented by this class.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._unsaved_entries = {}
        if not cls._instance and not cls._unsaved_entries:
            cls.read()
        return cls._instance

    def update(self, new_entries: Dict[str, cobib.database.Entry]) -> None:  # type: ignore
        """Updates the database with the given dictionary of entries.

        This function wraps `OrderedDict.update` and adds the labels of the changed entries to the
        dictionary of unsaved entries (`Database._unsaved_entries`). This will minimize IO access by
        only actually writing the unsaved entries in batches.

        Args:
            new_entries: the dictionary of labels mapping to entries which are to be written to the
                database.
        """
        for label in new_entries.keys():
            LOGGER.debug("Updating entry %s", label)
            Database._unsaved_entries[label] = label
        super().update(new_entries)

    def pop(self, label: str) -> cobib.database.Entry:  # type: ignore
        """Pops the entry pointed to by the given label.

        This function wraps `OrderedDict.pop` and adds the removed labels to the unsaved entries
        (`Database._unsaved_entries`). This will minimize IO access by only actually writing the
        unsaved entries in batches.

        Args:
            label: the label of the entry to be removed.

        Returns:
            The entry pointed to by the given label.
        """
        entry: cobib.database.Entry = super().pop(label)
        LOGGER.debug("Removing entry: %s", label)
        Database._unsaved_entries[label] = None
        return entry

    def rename(self, old_label: str, new_label: str) -> None:
        """Renames an entry label.

        This function performs no actual changes to the database. It merely registers a rename
        operation in `Database._unsaved_entries`.

        Args:
            old_label: the previous label.
            new_label: the new label.
        """
        LOGGER.debug("Renaming entry '%s' to '%s'.", old_label, new_label)
        Database._unsaved_entries[old_label] = new_label
        if new_label != old_label:
            # NOTE: this is not technically needed but the rename method is exploited during
            # database linting with "fake" renames in order to register entries for re-writing
            # during saving
            super().pop(old_label)

    def disambiguate_label(self, label: str, entry: cobib.database.Entry) -> str:
        """Disambiguate a given label to ensure it becomes unique.

        This function ensures that a label is unique by appending a configurable suffix to a label
        if it is already present in the database at runtime.

        Args:
            label: the label which to disambiguate.
            entry: the `Entry` to which this label belongs.

        Returns:
            A unique label.
        """
        if label not in self.keys() or self[label] == entry:
            return label

        LOGGER.warning("Label '%s' already exists in database. Running disambiguation.", label)
        separator, enumerator = config.database.format.label_suffix
        offset = 0
        while True:
            offset += 1
            new_label: str = label + separator + enumerator(offset)  # type: ignore[operator]
            if new_label not in self.keys():
                LOGGER.info("Found new unique label: %s", new_label)
                return new_label

    @classmethod
    def read(cls) -> None:
        """Reads the database file.

        The YAML database file pointed to by the configuration file is read in and parsed.
        This uses `cobib.parsers.YAMLParser` to parse the data.
        This function clears the contents of the singleton `Database` instance and resets
        `Database._unsaved_entries` to an empty dictionary. Thus, a call to this function
        *irreversibly* synchronizes the state of the runtime `Database` instance to the actually
        written contents of the database file on disc.
        """
        if cls._instance is None:
            cls()
            return
        _instance = cls._instance

        file = RelPath(config.database.file).path
        try:
            LOGGER.info("Loading database file: %s", file)
            # pylint: disable=import-outside-toplevel
            from cobib.parsers.yaml import YAMLParser

            _instance.clear()
            _instance.update(YAMLParser().parse(file))
        except FileNotFoundError:
            LOGGER.critical("The database file %s does not exist! Please run `cobib init`!", file)
            sys.exit(1)

        cls._unsaved_entries.clear()

    @classmethod
    def save(cls) -> None:
        """Saves all unsaved entries.

        This uses `cobib.parsers.YAMLParser` to save all entries in `Database._unsaved_entries` to
        disc. In doing so, this function preserves the order of the entries in the database file by
        overwriting changed entries in-place and appending new entries to the end of the file.

        The method of determining whether an entry was added, changed or removed is the following:
        1. we read in the current database as written to disc.
        2. we iterate all lines and determine the label of the entry we are currently on.
        3. if this label is not in `Database._unsaved_entries` we continue.
        4. Otherwise we query the runtime `Database` instance for the new contents of the unsaved
           (and therefore changed) entry and remove the label from `Database._unsaved_entries`.
        5. Using `Entry.save` and a `cobib.parsers.YAMLParser` we overwrite the previous lines of
           the changed entry.
        6. Finally, all labels still left in `Database._unsaved_entries` are newly added entries and
           can simply be appended to the file.

        In order to optimize performance and IO access, all of the above is done with a single call
        to `write`.
        """
        if cls._instance is None:
            cls()
        _instance = cast(Database, cls._instance)

        # pylint: disable=import-outside-toplevel
        from cobib.parsers.yaml import YAMLParser

        yml = YAMLParser()

        file = RelPath(config.database.file).path
        with open(file, "r", encoding="utf-8") as bib:
            lines = bib.readlines()

        label_regex = re.compile(r"^([^:]+):$")

        overwrite = False
        cur_label: str = ""
        buffer: List[str] = []
        for line in lines:
            try:
                matches = label_regex.match(line)
                if matches is None:
                    raise AttributeError
                new_label = matches.groups()[0]
                if new_label in cls._unsaved_entries:
                    LOGGER.debug('Entry "%s" found. Starting to replace lines.', new_label)
                    overwrite = True
                    cur_label = new_label
                    continue
            except AttributeError:
                pass
            if overwrite and line.startswith("..."):
                LOGGER.debug('Reached end of entry "%s".', cur_label)
                overwrite = False

                new_label = cls._unsaved_entries.pop(cur_label)
                entry = _instance.get(new_label, None)
                if entry:
                    LOGGER.debug('Writing modified entry "%s".', new_label)
                    entry_str = entry.save(parser=yml)
                    buffer.append("\n".join(entry_str.split("\n")[1:]))
                else:
                    # Entry has been deleted. Pop the previous `---` line.
                    LOGGER.debug('Deleting entry "%s".', new_label)
                    buffer.pop()
                # we pop `new_label` too, because in case of a rename it differs from `cur_label`
                if new_label is not None:
                    cls._unsaved_entries.pop(new_label, None)
            elif not overwrite:
                # keep previous line
                buffer.append(line)

        if cls._unsaved_entries:
            for label in cls._unsaved_entries.copy().values():
                if label is None:
                    # should never occur but we avoid a type exception
                    continue
                LOGGER.debug('Adding new entry "%s".', label)
                entry_str = _instance[label].save(parser=yml)
                buffer.append(entry_str)
                cls._unsaved_entries.pop(label)

        with open(file, "w", encoding="utf-8") as bib:
            for line in buffer:
                bib.write(line)



"""Tests for coBib's ZoteroImporter."""
# pylint: disable=unused-argument

import json
import logging
import tempfile
from itertools import zip_longest
from pathlib import Path
from typing import Dict, List

import pytest
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Entry
from cobib.importers import ZoteroImporter
from cobib.parsers import YAMLParser

from .. import get_resource
from .importer_test import ImporterTest

EXPECTED_DATABASE = get_resource("zotero_database.yaml", "importers")


class MockZoteroImporter(ZoteroImporter):
    """This class mocks the `ZoteroImporter` by providing fake OAuth authentication tokens."""

    @override
    @staticmethod
    def _get_authentication_tokens(no_cache: bool = False) -> Dict[str, str]:
        return {
            # NOTE: we are relying on the publicly accessible user `cobib` for testing purposes
            "UserID": "8608002",
        }


class TestZoteroImporter(ImporterTest):
    """Tests for coBib's ZoteroImporter."""

    def _assert_results(self, imported_entries: List[Entry]) -> None:
        """Common assertion utility method.

        Args:
            imported_entries: the list of entries to assert against the expected database.
        """
        parser = YAMLParser()

        imported_database = ""
        for entry in imported_entries:
            entry_str = parser.dump(entry)
            if entry_str is not None:
                imported_database += entry_str

        with open(EXPECTED_DATABASE, "r", encoding="utf-8") as expected:
            for line, truth in zip_longest(
                imported_database.splitlines(keepends=True), expected.readlines()
            ):
                assert line == truth

    @pytest.mark.asyncio
    async def test_fetch(self) -> None:
        """Test fetching entries from the Zotero API."""
        importer = MockZoteroImporter(skip_download=True)
        # NOTE: even though attachments are not accessible via public libraries, we explicitly skip
        # downloading them, just to be sure.
        imported_entries = await importer.fetch()

        self._assert_results(imported_entries)

    @pytest.mark.asyncio
    async def test_fetch_custom_user_id(self) -> None:
        """Test fetching with custom user ID."""
        imported_entries = await ZoteroImporter("--user-id", "8608002", "--no-cache").fetch()
        self._assert_results(imported_entries)

    @pytest.mark.asyncio
    async def test_cache(self) -> None:
        """Test caching behavior."""
        try:
            config.logging.cache = str(Path(tempfile.gettempdir()) / "cache")
            imported_entries = await ZoteroImporter(
                "--user-id", "8608002", skip_download=True
            ).fetch()
            self._assert_results(imported_entries)

            with open(config.logging.cache, "r", encoding="utf-8") as cache:
                cached_data = json.load(cache)
                assert "Zotero" in cached_data.keys()
                assert "UserID" in cached_data["Zotero"].keys()
                assert cached_data["Zotero"]["UserID"] == "8608002"
        finally:
            config.defaults()

    @pytest.mark.asyncio
    async def test_handle_argument_error(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test handling of ArgumentError.

        Args:
            caplog: the built-in pytest fixture.
        """
        try:
            await ZoteroImporter("--dummy").fetch()
        except SystemExit:
            pass
        for source, level, message in caplog.record_tuples:
            if ("cobib.importers.base_importer", logging.ERROR) == (
                source,
                level,
            ) and "Error: zotero: error:" in message:
                break
        else:
            pytest.fail("No Error logged from ArgumentParser.")

    @pytest.mark.asyncio
    async def test_event_pre_zotero_import(self) -> None:
        """Tests the PreZoteroImport event."""

        @Event.PreZoteroImport.subscribe
        def hook(importer: ZoteroImporter) -> None:
            importer.protected_url = importer.protected_url.replace("test", "8608002")

        assert Event.PreZoteroImport.validate()

        imported_entries = await ZoteroImporter("--user-id", "test", "--no-cache").fetch()

        self._assert_results(imported_entries)

    @pytest.mark.asyncio
    async def test_event_post_zotero_import(self) -> None:
        """Tests the PostZoteroImport event."""

        @Event.PostZoteroImport.subscribe
        def hook(importer: ZoteroImporter) -> None:
            importer.imported_entries = []

        assert Event.PostZoteroImport.validate()

        imported_entries = await MockZoteroImporter("--no-cache").fetch()
        assert imported_entries == []



"""coBib's Importer interface."""

from __future__ import annotations

import argparse
import logging
import sys
from abc import ABC, abstractmethod
from typing import List

from cobib.database import Entry
from cobib.ui.components import ArgumentParser as ArgumentParser

LOGGER = logging.getLogger(__name__)


class Importer(ABC):
    """The Importer interface.

    This interface should be implemented by all concrete importer implementations.
    """

    name = "base"
    """The importers `name` is used to register itself as an input argument to the
    `cobib.commands.import_.ImportCommand`."""

    argparser: ArgumentParser
    """Every importer has its own `argparse.ArgumentParser` which is used to parse the arguments
    provided to it."""

    def __init__(self, *args: str, skip_download: bool = False) -> None:
        """Initializes an importer instance.

        Args:
            *args: the sequence of additional importer arguments. These will be passed on to the
                `argparser` of this importer for further parsing.
            skip_download: whether or not to skip downloading of additional files such as attached
                PDF files or notes.
        """
        self.args: tuple[str, ...] = args
        """The raw provided importer arguments."""

        self.largs: argparse.Namespace = self.__class__._parse_args(args)
        """The parsed (local) arguments."""

        self.skip_download = skip_download
        """Whether or not to skip downloading of additional files such as attached PDF files or
        notes."""

    @classmethod
    @abstractmethod
    def init_argparser(cls) -> None:
        """Initializes this importer's `argparse.ArgumentParser`.

        This method needs to be overwritten by every subclass and handles the registration of all
        available importer arguments.
        """

    @classmethod
    def _get_argparser(cls) -> ArgumentParser:
        """Returns this importer's `argparse.ArgumentParser`.

        The reason for having this method is to handle the parser initialization such that it only
        needs to be done once.

        Returns:
            This importer's initialized `argparser` object.
        """
        if hasattr(cls, "argparser"):
            return cls.argparser

        cls.init_argparser()
        return cls.argparser

    @classmethod
    def _parse_args(cls, args: tuple[str, ...]) -> argparse.Namespace:
        """Parses the provided importer arguments.

        Args:
            args: the sequence of additional importer arguments provided to the importer upon
                initialization.

        Returns:
            The parsed arguments namespace.
        """
        try:
            largs = cls._get_argparser().parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            sys.exit(1)

        return largs

    @abstractmethod
    def fetch(self) -> List[Entry]:
        """Fetches the data from the source which this importer links to.

        Returns:
            A list of entries.
        """



"""Tests for coBib's InitCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import os
from datetime import datetime
from shutil import rmtree
from typing import TYPE_CHECKING, Any, Type

import pytest
from typing_extensions import override

from cobib.commands import InitCommand
from cobib.config import Event, config

from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestInitCommand(CommandTest):
    """Tests for coBib's InitCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return InitCommand

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False, "database": False}],
            [{"git": True, "database": False}],
            [{"git": True, "database": True}],
        ],
        indirect=["setup"],
    )
    @pytest.mark.parametrize(["safe"], [[False], [True]])
    def test_command(self, setup: Any, safe: bool) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            safe: whether the safety of the `InitCommand` should be checked.
        """
        if safe:
            # fill database file
            with open(config.database.file, "w", encoding="utf-8") as file:
                file.write("test")
        # store current time
        now = float(datetime.now().timestamp())
        # try running init
        if setup["git"]:
            InitCommand("--git").execute()
        else:
            InitCommand().execute()
        if safe:
            # check database file still contains 'test'
            with open(config.database.file, "r", encoding="utf-8") as file:
                assert file.read() == "test"
        else:
            # check creation time of temporary database file
            ctime = os.stat(config.database.file).st_ctime
            # assert these times are close
            assert ctime - now < 0.1 or now - ctime < 0.1
        if setup["git"] and not setup["database"]:
            # check creation time of temporary database git folder
            ctime = os.stat(self.COBIB_TEST_DIR_GIT).st_ctime
            # assert these times are close
            assert ctime - now < 0.1 or now - ctime < 0.1
            # and assert that it is indeed a folder
            assert self.COBIB_TEST_DIR_GIT.is_dir()
            # assert the git commit message
            self.assert_git_commit_message("init", {"git": True})

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False, "database": False}],
        ],
        indirect=["setup"],
    )
    def test_warn_insufficient_config(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning in case of insufficient config.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        try:
            # store current time
            now = float(datetime.now().timestamp())
            # try running init
            InitCommand("--git").execute()

            # assert warning is printed
            assert (
                "cobib.commands.init",
                30,
                "You are about to initialize the git tracking of your database, but this will only "
                "have effect if you also enable the DATABASE/git setting in your configuration "
                "file!",
            ) in caplog.record_tuples
            # now assert that the command did everything as usual though

            # check creation time of temporary database file
            ctime = os.stat(config.database.file).st_ctime
            # assert these times are close
            assert ctime - now < 0.1 or now - ctime < 0.1
            # check creation time of temporary database git folder
            ctime = os.stat(self.COBIB_TEST_DIR_GIT).st_ctime
            # assert these times are close
            assert ctime - now < 0.1 or now - ctime < 0.1
            # and assert that it is indeed a folder
            assert self.COBIB_TEST_DIR_GIT.is_dir()
            # assert the git commit message
            self.assert_git_commit_message("init", {"git": True})
        finally:
            # clean up file system
            rmtree(self.COBIB_TEST_DIR_GIT)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False, "database": False}],
        ],
        indirect=["setup"],
    )
    # other variants are already covered by test_command
    async def test_cmdline(self, setup: Any, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
        """
        # store current time
        now = float(datetime.now().timestamp())
        # try calling init
        await self.run_module(monkeypatch, "main", ["cobib", "init"])
        # try running init
        # check creation time of temporary database file
        ctime = os.stat(config.database.file).st_ctime
        # assert these times are close
        assert ctime - now < 0.1 or now - ctime < 0.1

    def test_event_pre_init_command(self, setup: Any) -> None:
        """Tests the PreInitCommand event."""

        @Event.PreInitCommand.subscribe
        def hook(command: InitCommand) -> None:
            command.largs.git = True

        assert Event.PreInitCommand.validate()

        InitCommand().execute()

        self.assert_git_commit_message("init", {"git": True})

    def test_event_post_init_command(self, setup: Any) -> None:
        """Tests the PostInitCommand event."""

        @Event.PostInitCommand.subscribe
        def hook(command: InitCommand) -> None:
            os.remove(command.file)

        assert Event.PostInitCommand.validate()

        InitCommand().execute()

        assert not os.path.exists(self.COBIB_TEST_DIR_GIT / "literature.yaml")



"""coBib's DOI parser.

This parser is capable of generating `cobib.database.Entry` instances from a given DOI.
It gathers the BibTex-encoded data from https://doi.org/ and parses it directly using the
`cobib.parsers.bibtex.BibtexParser`.

Since v3.2.0 coBib will also attempt to download the PDF version of the new entry. You can
configure the default download location via
`cobib.config.config.FileDownloaderConfig.default_location`.
Since in general the PDF may not be freely available, your mileage with this feature may vary. Until
coBib supports internal proxy configurations, make sure you are logged in to a VPN for the smoothest
experience with closed-source journals.
Furthermore, you should look into the `cobib.config.config.FileDownloaderConfig.url_map` setting,
through which you tell coBib how to map from journal landing page URLs to the corresponding PDF
URLs. For more information check out `cobib.config.example` and the man-page.

Since v3.3.0 this parser even supports URLs from which a DOI can be extracted directly.

The parser is registered under the `-d` and `--doi` command-line arguments of the
`cobib.commands.add.AddCommand`.

The following documentation is mostly inherited from the abstract interface
`cobib.parsers.base_parser`.
"""

import logging
import re
from collections import OrderedDict
from typing import Dict

import requests
from typing_extensions import override

from cobib.config import Event
from cobib.database import Entry

from .base_parser import Parser
from .bibtex import BibtexParser

LOGGER = logging.getLogger(__name__)

DOI_URL = "https://doi.org/"
"""The DOI 'API' URL."""
DOI_HEADER = {"Accept": "application/x-bibtex"}
"""The DOI 'API' header taken from [here](https://crosscite.org/docs.html)."""
DOI_REGEX = r'(10\.[0-9a-zA-Z]+\/(?:(?!["&\'\?])\S)+)\b'
"""A regex pattern used to match valid DOIs."""


class DOIParser(Parser):
    """The DOI Parser."""

    name = "doi"

    @override
    def parse(self, string: str) -> Dict[str, Entry]:
        string = Event.PreDOIParse.fire(string) or string

        try:
            match = re.search(DOI_REGEX, string)
            if match is None:
                raise AssertionError
        except AssertionError:
            msg = f"'{string}' is not a valid DOI."
            LOGGER.warning(msg)
            return OrderedDict()
        doi = match.group(1)
        LOGGER.info("Gathering BibTex data for DOI: %s.", doi)
        try:
            page = requests.get(DOI_URL + doi, headers=DOI_HEADER, timeout=10)
            # this assumes that the doi.org page redirects to the correct journal's landing page
            redirected_url = requests.head(DOI_URL + doi, timeout=1).headers["Location"]
        except requests.exceptions.RequestException as err:
            LOGGER.error("An Exception occurred while trying to query the DOI: %s.", doi)
            LOGGER.error(err)
            return OrderedDict()
        bib = BibtexParser().parse(page.text)
        for entry in bib.values():
            entry.data["_download"] = redirected_url

        Event.PostDOIParse.fire(bib)

        return bib

    def dump(self, entry: Entry) -> None:
        """We cannot dump a generic entry as a DOI."""
        LOGGER.error("Cannot dump an entry as a DOI.")



"""Tests for coBib's OpenCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Type

import pytest
from typing_extensions import override

from cobib.commands import OpenCommand
from cobib.config import Event, config
from cobib.database import Database

from .. import MockStdin, get_resource
from .command_test import CommandTest

if TYPE_CHECKING:
    import _pytest.fixtures

    import cobib.commands


class TestOpenCommand(CommandTest):
    """Tests for coBib's OpenCommand."""

    # Note: we can hard-code the `/tmp` path here, because we never really create these files.
    # We just need some absolute paths to test against.
    TMP_FILE_A = "/tmp/a.txt"
    TMP_FILE_B = "/tmp/b.txt"

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return OpenCommand

    @pytest.fixture
    def post_setup(
        self, monkeypatch: pytest.MonkeyPatch, request: _pytest.fixtures.SubRequest
    ) -> Generator[Dict[str, Any], None, None]:
        """Additional setup instructions.

        Args:
            monkeypatch: the built-in pytest fixture.
            request: a pytest sub-request providing access to nested parameters.

        Yields:
            The internally used parameters for potential later re-use during the actual test.
        """
        if not hasattr(request, "param"):
            # use default settings
            request.param = {"stdin_list": None, "multi_file": True}

        if request.param.get("multi_file", True):
            with open(
                get_resource("example_multi_file_entry.yaml", "commands"), "r", encoding="utf-8"
            ) as multi_file_entry:
                with open(config.database.file, "a", encoding="utf-8") as database:
                    database.write(multi_file_entry.read())
            Database().read()

        monkeypatch.setattr("sys.stdin", MockStdin(request.param.get("stdin_list", None)))

        yield request.param

    def _assert(  # type: ignore
        self, output: List[str], logs: Optional[List[Tuple[str, int, str]]] = None, **kwargs
    ) -> None:
        """Common assertion utility method.

        Args:
            output: the list of lines printed to `sys.stdout`.
            logs: the list of logged messages.
            kwargs: additional test-specific keyword arguments.
        """
        if not kwargs.get("multi_file", True):
            expected_log = [
                ("cobib.commands.open", 10, "Starting Open command."),
                (
                    "cobib.commands.open",
                    10,
                    'Parsing "http://www-cs-faculty.stanford.edu/\\~{}uno/abcde.html" for URLs.',
                ),
                (
                    "cobib.commands.open",
                    10,
                    'Opening "http://www-cs-faculty.stanford.edu/\\~{}uno/abcde.html" with cat.',
                ),
            ]
            if logs is not None:
                assert logs == expected_log
        else:
            expected_out = [
                "  1: [file] " + self.TMP_FILE_A,
                "  2: [file] " + self.TMP_FILE_B,
                "  3: [url] https://www.duckduckgo.com",
                "  4: [url] https://www.google.com",
                "[all,help,cancel]: ",
            ]

            expected_log = [
                ("cobib.commands.open", 10, "Starting Open command."),
                ("cobib.commands.open", 10, 'Parsing "' + self.TMP_FILE_A + '" for URLs.'),
                ("cobib.commands.open", 10, 'Parsing "' + self.TMP_FILE_B + '" for URLs.'),
                ("cobib.commands.open", 10, 'Parsing "https://www.duckduckgo.com" for URLs.'),
                ("cobib.commands.open", 10, 'Parsing "https://www.google.com" for URLs.'),
            ]

            stdin_list = kwargs.get("stdin_list", [])
            extra_logs = None
            if "help" in stdin_list:
                expected_out += expected_out.copy()
                expected_out[4] += "Multiple targets were found. You may select the following:"
                extra_out = [
                    "  1. an individual URL number",
                    "  2. a target type (provided in '[...]')",
                    "  3. 'all'",
                    "  4. or 'cancel' to abort the command",
                ]
                for line in reversed(extra_out):
                    expected_out.insert(5, line)

                expected_log.append(("cobib.commands.open", 10, "User requested help."))
                expected_log.append(("cobib.commands.open", 30, "User aborted open command."))
            elif "cancel" in stdin_list:
                expected_log.append(("cobib.commands.open", 30, "User aborted open command."))
            elif "all" in stdin_list:
                extra_logs = [
                    ("cobib.commands.open", 10, "User selected all urls."),
                    ("cobib.commands.open", 10, 'Opening "' + self.TMP_FILE_A + '" with cat.'),
                    ("cobib.commands.open", 10, 'Opening "' + self.TMP_FILE_B + '" with cat.'),
                    ("cobib.commands.open", 10, 'Opening "https://www.duckduckgo.com" with cat.'),
                    ("cobib.commands.open", 10, 'Opening "https://www.google.com" with cat.'),
                ]
            elif "url" in stdin_list:
                extra_logs = [
                    ("cobib.commands.open", 10, "User selected the url set of urls."),
                    ("cobib.commands.open", 10, 'Opening "https://www.duckduckgo.com" with cat.'),
                    ("cobib.commands.open", 10, 'Opening "https://www.google.com" with cat.'),
                ]
            elif "1" in stdin_list:
                extra_logs = [
                    ("cobib.commands.open", 10, "User selected url 1"),
                    ("cobib.commands.open", 10, 'Opening "' + self.TMP_FILE_A + '" with cat.'),
                ]

            if extra_logs is not None:
                expected_log.extend(extra_logs)

            for line, truth in zip(output, expected_out):
                assert line == truth
            if logs is not None:
                assert logs == expected_log

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["args", "post_setup"],
        [
            [["knuthwebsite"], {"multi_file": False}],
            [["example_multi_file_entry"], {"multi_file": True, "stdin_list": ["cancel"]}],
            [["example_multi_file_entry"], {"multi_file": True, "stdin_list": ["help", "cancel"]}],
            [["example_multi_file_entry"], {"multi_file": True, "stdin_list": ["all"]}],
            [["example_multi_file_entry"], {"multi_file": True, "stdin_list": ["url"]}],
            [["example_multi_file_entry"], {"multi_file": True, "stdin_list": ["1"]}],
        ],
        indirect=["post_setup"],
    )
    async def test_command(  # pylint: disable=invalid-overridden-method
        self,
        setup: Any,
        post_setup: Any,
        caplog: pytest.LogCaptureFixture,
        capsys: pytest.CaptureFixture[str],
        args: List[str],
    ) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            post_setup: an additional setup fixture.
            caplog: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
            args: the arguments to pass to the command.
        """
        await OpenCommand(*args).execute()

        true_log = [rec for rec in caplog.record_tuples if rec[0] == "cobib.commands.open"]
        true_out = capsys.readouterr().out.split("\n")

        self._assert(true_out, true_log, **post_setup)

    @pytest.mark.asyncio
    async def test_warning_missing_label(
        self, setup: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test warning for missing label.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        await OpenCommand("dummy").execute()
        assert (
            "cobib.commands.open",
            30,
            "No entry with the label 'dummy' could be found.",
        ) in caplog.record_tuples

    @pytest.mark.asyncio
    async def test_warning_nothing_to_open(
        self, setup: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test warning for label with nothing to open.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        await OpenCommand("einstein").execute()
        assert (
            "cobib.commands.open",
            30,
            "The entry 'einstein' has no actionable field associated with it.",
        ) in caplog.record_tuples

    @pytest.mark.asyncio
    async def test_config_openable_fields(
        self,
        setup: Any,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Test the `config.commands.open.fields` setting.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
        """
        config.commands.open.fields = ["note"]

        with open(
            get_resource("example_multi_file_entry.yaml", "commands"), "r", encoding="utf-8"
        ) as multi_file_entry:
            with open(config.database.file, "a", encoding="utf-8") as database:
                for line in multi_file_entry.readlines():
                    if line == "  file:\n":
                        database.write("  note:\n")
                    else:
                        database.write(line)

        Database().read()

        monkeypatch.setattr("sys.stdin", MockStdin(["cancel"]))

        await OpenCommand("example_multi_file_entry").execute()

        true_out = capsys.readouterr().out.split("\n")

        expected_out = [
            "  1: [note] " + self.TMP_FILE_A,
            "  2: [note] " + self.TMP_FILE_B,
            "[all,help,cancel]: ",
        ]

        for line, truth in zip(true_out, expected_out):
            assert line == truth

    @pytest.mark.asyncio
    async def test_open_non_list_field(
        self,
        setup: Any,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Test opening of non-list type fields.

        This is a regression test against https://gitlab.com/cobib/cobib/-/issues/100

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            caplog: the built-in pytest fixture.
        """
        config.commands.open.fields = ["note"]

        with open(
            get_resource("example_multi_file_entry.yaml", "commands"), "r", encoding="utf-8"
        ) as multi_file_entry:
            with open(config.database.file, "a", encoding="utf-8") as database:
                for line in multi_file_entry.readlines():
                    if line == "  file:\n":
                        database.write("  note: /tmp/a.txt\n")
                        database.write("...\n")
                        break
                    database.write(line)

        Database().read()

        monkeypatch.setattr("sys.stdin", MockStdin())

        await OpenCommand("example_multi_file_entry").execute()

        true_log = [rec for rec in caplog.record_tuples if rec[0] == "cobib.commands.open"]

        expected_log = [
            ("cobib.commands.open", 10, "Starting Open command."),
            (
                "cobib.commands.open",
                10,
                'Parsing "/tmp/a.txt" for URLs.',
            ),
            (
                "cobib.commands.open",
                10,
                'Opening "/tmp/a.txt" with cat.',
            ),
        ]
        assert true_log == expected_log

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["post_setup"],
        [
            [{"multi_file": False}],
        ],
        indirect=["post_setup"],
    )
    async def test_cmdline(
        self,
        setup: Any,
        post_setup: Any,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            post_setup: an additional setup fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
        """
        await self.run_module(monkeypatch, "main", ["cobib", "open", "knuthwebsite"])

        true_out = capsys.readouterr().out.split("\n")

        self._assert(true_out, logs=None, **post_setup)

    @pytest.mark.asyncio
    async def test_event_pre_open_command(
        self,
        setup: Any,
        post_setup: Any,
        caplog: pytest.LogCaptureFixture,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Tests the PreOpenCommand event."""

        @Event.PreOpenCommand.subscribe
        def hook(command: OpenCommand) -> None:
            command.largs.labels = ["knuthwebsite"]

        assert Event.PreOpenCommand.validate()

        await OpenCommand("einstein").execute()

        true_log = [rec for rec in caplog.record_tuples if rec[0] == "cobib.commands.open"]
        true_out = capsys.readouterr().out.split("\n")

        self._assert(true_out, true_log, multi_file=False)

    @pytest.mark.asyncio
    async def test_event_post_open_command(
        self,
        setup: Any,
        post_setup: Any,
        caplog: pytest.LogCaptureFixture,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Tests the PostOpenCommand event."""

        @Event.PostOpenCommand.subscribe
        def hook(command: OpenCommand) -> None:
            print(command.largs.labels, file=sys.stderr)

        assert Event.PostOpenCommand.validate()

        await OpenCommand("knuthwebsite").execute()

        true_log = [rec for rec in caplog.record_tuples if rec[0] == "cobib.commands.open"]
        outerr = capsys.readouterr()
        true_out = outerr.out.split("\n")

        self._assert(true_out, true_log, multi_file=False)

        assert outerr.err == "['knuthwebsite']\n"



"""coBib's UIs.

coBib is designed with two user interfaces in mind:
- a command-line interface (short: CLI)
- a terminal user interface (short: TUI)

Both of these, as well as some other common user-interface utilities, are part of this module.

In addition, coBib also has some shell helpers (see `cobib.utils.shell_helper`) which are exposed
via the separate `ShellHelper` interface (generally speaking those commands need to be prefixed with
an underscore, for example: `cobib _example_config`).
"""

from .cli import CLI as CLI
from .shell_helper import ShellHelper as ShellHelper
from .ui import UI as UI



"""coBib's shell helper interface.

This class provides access to the commands implemented in the `cobib.utils.shell_helper` module.
A user can access those from the command-line by prefixing the various function names of that module
with an underscore, for example: `cobib _example_config`.
"""

import argparse
import inspect
import logging
from typing import Any

from typing_extensions import override

from cobib.ui.ui import UI
from cobib.utils import shell_helper

LOGGER = logging.getLogger(__name__)


class ShellHelper(UI):
    """The shell-helper interface class.

    In addition to the global arguments documented by the base class, the following are supported:

      * `helper`: a single position argument indicating the name of the shell-helper utility
        function to run (prefixed with a single underscore, for example `_example_config`).
    """

    @override
    def add_extra_parser_arguments(self) -> None:
        available_helpers = [
            "_" + m[0] for m in inspect.getmembers(shell_helper) if inspect.isfunction(m[1])
        ]
        self.parser.add_argument(
            "helper", help="shell helper to be called", choices=available_helpers
        )
        self.parser.add_argument("args", nargs=argparse.REMAINDER)

    @override
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()

        self.init_argument_parser(description="Process shell helper call")

    @override
    def run(self) -> None:
        arguments = self.parse_args()

        helper = getattr(shell_helper, arguments.helper.strip("_"))
        # any shell helper function will return a list of the requested items
        for item in helper(*arguments.args):
            print(item)



"""coBib's logging module.

This module provides utility methods to set up logging to different handlers.
"""

import logging
import logging.handlers
import sys
from importlib import metadata
from pathlib import Path
from typing import List, Optional, Union

from rich.console import ConsoleRenderable, Group
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text
from typing_extensions import override

from .rel_path import RelPath


class _StderrHandler(logging.StreamHandler):  # type: ignore[type-arg]
    """A logging handler hard-coded to `sys.stderr`.

    The reason for explicitly deriving this class, is that Python's `logging.StreamHandler` does not
    respect stream redirection. However, for coBib's TUI this is an important requirement which can
    be achieved by a runtime check of the stream during the `emit` method.
    """

    @override
    def emit(self, record: logging.LogRecord) -> None:
        self.stream = sys.stderr
        super().emit(record)


def get_stream_handler(
    level: int = logging.WARNING,
) -> logging.StreamHandler:  # type: ignore[type-arg]
    """Returns a basic StreamHandler logging to `sys.stderr`.

    Args:
        level: the logging level of this handler.
    """
    formatter = logging.Formatter(fmt="[%(levelname)s] %(message)s")

    handler = _StderrHandler()
    handler.setLevel(level)
    handler.setFormatter(formatter)

    return handler


def get_file_handler(
    level: Union[str, int] = "INFO", logfile: Optional[Union[str, Path]] = None
) -> logging.handlers.RotatingFileHandler:
    """Returns a `RotatingFileHandler`.

    Args:
        level: the handler's logging level.
        logfile: the output path for the log file.
    """
    if logfile is None:
        # pylint: disable=import-outside-toplevel
        from cobib.config import config

        logfile = config.logging.logfile

    path = RelPath(logfile).path
    path.parent.mkdir(parents=True, exist_ok=True)

    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(name)s %(funcName)s:%(lineno)d %(message)s"
    )

    handler = logging.handlers.RotatingFileHandler(
        filename=path,
        maxBytes=10485760,  # 10 Megabytes
        backupCount=10,
    )
    handler.setLevel(level)
    handler.setFormatter(formatter)

    return handler


def print_changelog(version: str, cached_version_path: Optional[str]) -> Optional[Panel]:
    """Generates a `rich.Panel` to display the changelog since the last displayed version.

    This function prints the contents of the CHANGELOG (extracted from the package metadata)
    between the current version (`version`) and the latest cached version (extracted from the
    provided file path).

    Args:
        version: the currently running version of coBib.
        cached_version_path: the path to the file which caches the previously executed version of
            coBib. If `None`, this method exits early (thereby silencing this feature).

    Returns:
        An optional `rich.Panel` with the rendered Markdown.
    """
    if cached_version_path is None:
        return None

    _cached_version_path = RelPath(cached_version_path).path
    if not _cached_version_path.parent.exists():
        _cached_version_path.parent.mkdir(parents=True)

    try:
        current_version = version[: version.index("+")]
    except ValueError:
        current_version = version

    cached_version = None
    try:
        with open(_cached_version_path, "r", encoding="utf-8") as version_file:
            cached_version = version_file.read().strip()
    except FileNotFoundError:
        pass

    if current_version == cached_version:
        return None

    with open(_cached_version_path, "w", encoding="utf-8") as version_file:
        version_file.write(current_version)

    groups: List[ConsoleRenderable] = []
    groups.append(
        Text(
            "Hi there! It looks like you have updated coBib; here is what's new:",
            style="bold yellow",
        )
    )

    description = str(metadata.metadata("cobib").get("description"))

    lines: List[str] = []
    started = False
    for line in description.splitlines():
        line = line.strip()
        if line.startswith(f"## [{current_version}]"):
            started = True
        elif line.startswith(f"## [{cached_version}]"):
            break

        if started:
            lines.append(line)

    groups.append(Markdown("\n".join(lines)))

    return Panel(Group(*groups))



# mypy: warn-unused-ignores=False
"""coBib's YAML parser.

This parser leverages the [`ruamel.yaml`](https://pypi.org/project/ruamel.yaml/) library to convert
between `cobib.database.Entry` instances and YAML representations of their `dict`-like data
structure.

The parser is registered under the `-y` and `--yaml` command-line arguments of the
`cobib.commands.add.AddCommand`.

The following documentation is mostly inherited from the abstract interface
`cobib.parsers.base_parser`.
"""

import io
import logging
import sys
from collections import OrderedDict
from pathlib import Path
from typing import IO, Dict, Optional, Union

from rich.console import Console
from rich.progress import track
from ruamel import yaml
from typing_extensions import override

from cobib.config import Event, config
from cobib.database.entry import Entry

from .base_parser import Parser

LOGGER = logging.getLogger(__name__)


class YAMLParser(Parser):
    """The YAML Parser."""

    name = "yaml"

    _yaml: Optional[yaml.YAML] = None

    def __init__(self) -> None:  # pylint: disable=C0116
        # noqa: D107
        if YAMLParser._yaml is None:
            # we need to lazily construct this in order to be able to respect the config setting
            YAMLParser._yaml = yaml.YAML(typ="safe", pure=not config.parsers.yaml.use_c_lib_yaml)
            YAMLParser._yaml.explicit_start = True  # type: ignore[assignment]
            YAMLParser._yaml.explicit_end = True  # type: ignore[assignment]
            YAMLParser._yaml.default_flow_style = False

    @override
    def parse(self, string: Union[str, Path]) -> Dict[str, Entry]:
        string = Event.PreYAMLParse.fire(string) or string

        bib = OrderedDict()
        LOGGER.debug("Loading YAML data from file: %s.", string)
        try:
            stream: IO = io.StringIO(Path(string))  # type: ignore[arg-type,type-arg]
        except TypeError:
            try:
                stream = open(string, "r", encoding="utf-8")  # pylint: disable=consider-using-with
            except FileNotFoundError as exc:
                raise exc
        for entry in track(
            self._yaml.load_all(stream),  # type: ignore[union-attr]
            description="Reading database...",
            transient=True,
            console=Console(file=sys.stderr),  # TODO: do not hard-code this
        ):
            for label, data in entry.items():
                bib[label] = Entry(label, data)
        stream.close()

        Event.PostYAMLParse.fire(bib)

        return bib

    @override
    def dump(self, entry: Entry) -> Optional[str]:
        Event.PreYAMLDump.fire(entry)

        LOGGER.debug("Converting entry %s to YAML format.", entry.label)
        stream = io.StringIO()
        self._yaml.dump(  # type: ignore[union-attr]
            {entry.label: dict(sorted(entry.data.items()))}, stream=stream
        )
        string = stream.getvalue()

        string = Event.PostYAMLDump.fire(string) or string

        return string



"""coBib's diff renderer utility."""

from __future__ import annotations

import difflib
import re
from collections import namedtuple
from dataclasses import dataclass, field

from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table

Coordinate = namedtuple("Coordinate", ["row", "col"])


@dataclass
class Side:
    """A simple dataclass for configuring the different sides of a diff view."""

    name: str
    """The name of the side."""
    background: str
    """What background color to use for highlighted differences on this side."""
    foreground: str
    """What foreground color to use for highlighted differences on this side."""


LEFT = Side("a", "#300000", "red")
"""The configured left side of a diff view."""
RIGHT = Side("b", "#002000", "green")
"""The configured right side of a diff view."""


@dataclass
class DiffData:
    """A simple dataclass for holding the difference data."""

    lines: list[str]
    """The total lines being compared."""

    highlight_lines: set[int] = field(default_factory=set)
    """The indices of the lines to be highlighted because they differ."""

    style_ranges: list[tuple[Style, Coordinate, Coordinate]] = field(default_factory=list)
    """A list of stylized ranges to further highlight differences. This is a list of tuples, each of
    which contains three elements:

    1. the style to be applied to this range
    2. the initial `(row, col)` coordinate
    3. the final `(row, col)` coordinate
    """

    current_block: list[str] = field(default_factory=list)
    """The current list of lines being compared."""

    current_offset: int = 0
    """The current index offset of the lines being compared."""

    def _get_cumulative_count(self) -> int:
        """Returns the cumulative number of characters up to the current offset."""
        return sum(len(line) for line in self.lines[: self.current_offset - 1])

    def get_row_col_from_pos(self, position: int) -> Coordinate:
        """Converts a character position into a `(row, col)` coordinate.

        Args:
            position: the absolute character position.

        Returns:
            The `(row, col)` coordinate.
        """
        position += self._get_cumulative_count()
        col, row = -1, 0
        cumulative_count = 0
        for line in self.lines:
            cumulative_count += len(line)
            if cumulative_count > position:
                col = position - (cumulative_count - len(line))
                break
            row += 1
        return Coordinate(row + 1, col)

    def stylize_matches(self, matches: list[difflib.Match], side: Side) -> None:
        """Processes the provided matches and registers the styles in `self`.

        Args:
            matches: the list of difference matches to be highlighted.
            side: the styling data for a specific side.
        """
        pos = 0
        for match in matches:
            row1, col1 = self.get_row_col_from_pos(pos)
            row2, col2 = self.get_row_col_from_pos(getattr(match, side.name))
            pos = getattr(match, side.name) + match.size

            for row in range(row1, row2 + 1 * (match.size > 0)):
                self.highlight_lines.add(row)

                self.style_ranges.insert(
                    0, (Style(bgcolor=side.background), Coordinate(row, 0), Coordinate(row + 1, 0))
                )

            self.style_ranges.append(
                (Style(bgcolor=side.foreground), Coordinate(row1, col1), Coordinate(row2, col2))
            )

    def render(self, lexer: str = "text") -> Syntax:
        """Renders the difference data as a `rich` renderable.

        Args:
            lexer: the syntax lexing to apply.

        Returns:
            A `rich` renderable.
        """
        syntax = Syntax(
            "".join(self.lines),
            lexer,
            background_color="default",
            highlight_lines=self.highlight_lines,
            line_numbers=True,
            word_wrap=True,
        )

        for style in self.style_ranges:
            syntax.stylize_range(*style)

        return syntax


class Differ:
    """A utility class to compare two strings."""

    _unified_diff_header = re.compile(r"@@ -(\d+),?(\d*) \+(\d+),?(\d*) @@")
    """A regex to detect the headers of a `unified_diff` output."""

    def __init__(self, left: str, right: str) -> None:
        """Initializes a `Differ` instance.

        Args:
            left: the left string.
            right: the right string.
        """
        self.left = DiffData(left.splitlines(keepends=True))
        self.right = DiffData(right.splitlines(keepends=True))

    def _sequence_matcher(self) -> None:
        """A utility method to process the `difflib.SequenceMatcher` results."""
        matches = difflib.SequenceMatcher(
            a="".join(self.left.current_block),
            b="".join(self.right.current_block),
        ).get_matching_blocks()

        self.left.stylize_matches(matches, LEFT)
        self.right.stylize_matches(matches, RIGHT)

        self.left.current_block = []
        self.right.current_block = []

    def _parse_unified_diff_header(self, line: str) -> None:
        """A utility to process a `difflib.unified_diff` header."""
        result = Differ._unified_diff_header.search(line)

        if result is not None:
            offset_left, _, offset_right, _ = result.groups()

            self.left.current_offset = int(offset_left)
            self.right.current_offset = int(offset_right)

    def compute(self) -> None:
        """Computes the difference between the two strings.

        This combines the information gathered from `difflib.unified_diff` and additionally compares
        each block with a `difflib.SequenceMatcher`.
        """
        udiff = difflib.unified_diff(
            self.left.lines,
            self.right.lines,
            n=0,
            lineterm="",
        )

        for line in udiff:
            if line[:2] == "@@":
                self._sequence_matcher()
                self._parse_unified_diff_header(line)

            elif line[:1] == "-" and line != "--- ":
                self.left.current_block.append(line[1:])

            elif line[:1] == "+" and line != "+++ ":
                self.right.current_block.append(line[1:])

        self._sequence_matcher()

    def render(self, lexer: str = "text") -> Table:
        """Renders the compared strings in a side-by-side view.

        Args:
            lexer: the syntax lexing to apply.

        Returns:
            A `rich` renderable.
        """
        table = Table(box=None, show_header=False)
        table.add_row(self.left.render(lexer), self.right.render(lexer))
        return table



"""coBib's Open command.

This command can be used to open associated file(s) of one (or multiple) entries.
```
cobib open <label 1> [<label 2> ...]
```

The keys of `cobib.database.Entry.data` which are queried for paths or URL strings can be configured
via the `cobib.config.config.OpenCommandConfig.fields` setting (defaulting to `["file", "url"]`).
If one such string is found, it is automatically opened with the program configured by
`cobib.config.config.OpenCommandConfig.command`.
If multiple matches are found, the user will be presented with a menu to choose one or multiple
matches.

This menu will look similar to the following after querying for `help`:
```
Multiple targets were found. You may select the following:
  1. an individual URL number
  2. a target type (provided in '[...]')
  3. 'all'
  4. or 'cancel' to abort the command

  1: [file] /path/to/a/file.pdf
  2: [file] /path/to/another/file.pdf
  3: [url] https://example.org/
[all,help,cancel]:
```

With the above options, here is what will happen depending on the users choice:
* `1`, `2`, or `3`: will open the respective file or URL.
* `file` or `url`: will open the respective group.
* `all`: will open all matches.
* `help`: will print the detailed help-menu again.
* `cancel`: will abort the command.

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `o` key.
"""

from __future__ import annotations

import logging
import os
import subprocess
import warnings
from collections import defaultdict
from functools import wraps
from typing import Callable, cast
from urllib.parse import ParseResult, urlparse

from rich.console import Console
from rich.prompt import InvalidResponse, Prompt, PromptBase, PromptType
from rich.text import Text
from textual.app import App
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class OpenCommand(Command):
    """The Open Command.

    This command can parse the following arguments:

        * `labels`: one (or multiple) labels of the entries to be opened.
    """

    name = "open"

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="open", description="Open subcommand parser.")
        parser.add_argument("labels", type=str, nargs="+", help="labels of the entries")
        cls.argparser = parser

    # TODO: can we make the implementation cleaner and avoid the type ignore comment below?
    @override
    async def execute(self) -> None:  # type: ignore[override]
        # pylint: disable=invalid-overridden-method
        LOGGER.debug("Starting Open command.")

        Event.PreOpenCommand.fire(self)

        bib = Database()

        for label in self.largs.labels:
            things_to_open: dict[str, list[ParseResult]] = defaultdict(list)
            count = 0
            # first: find all possible things to open
            try:
                entry = bib[label]
                for field in config.commands.open.fields:
                    if field in entry.data.keys() and entry.data[field]:
                        value = entry.data[field]
                        if not isinstance(value, list):
                            value = [value]
                        for val in value:
                            val = val.strip()
                            LOGGER.debug('Parsing "%s" for URLs.', val)
                            things_to_open[field] += [urlparse(val)]
                            count += 1
            except KeyError:
                msg = f"No entry with the label '{label}' could be found."
                LOGGER.warning(msg)
                continue

            # if there are none, skip current label
            if not things_to_open:
                msg = f"The entry '{label}' has no actionable field associated with it."
                LOGGER.warning(msg)
                continue

            if count == 1:
                # we found a single URL to open
                self._open_url(list(things_to_open.values())[0][0])
            else:
                # we query the user what to do
                idx = 0
                url_list: list[ParseResult] = []
                prompt_text = Text()
                choices = ["all"] + config.commands.open.fields + ["help", "cancel"]

                # print formatted list of available URLs
                for field, urls in things_to_open.items():
                    for url in urls:
                        idx += 1
                        url_list.append(url)
                        choices.append(str(idx))
                        prompt_text.append(f"{idx:3}", "prompt.choices")
                        prompt_text.append(": [")
                        prompt_text.append(field, "prompt.choices")
                        prompt_text.append(f"] {url.geturl()}\n")
                prompt_text.append("[all,help,cancel]", "prompt.choices")

                # pylint: disable=line-too-long
                self.prompt.process_response = self._wrap_prompt_process_response(  # type: ignore[method-assign]
                    self.prompt.process_response  # type: ignore[assignment]
                )
                if self.prompt is not Prompt:
                    choice = await self.prompt.ask(  # type: ignore[call-overload]
                        prompt_text,
                        choices=choices,
                        show_choices=False,
                        console=cast(App[None], self.console),
                    )
                else:
                    choice = self.prompt.ask(
                        prompt_text,
                        choices=choices,
                        show_choices=False,
                        console=cast(Console, self.console),
                    )

                self.prompt.process_response = (  # type: ignore[method-assign]
                    self.prompt.process_response.__wrapped__  # type: ignore[attr-defined]
                )

                if choice == "cancel":
                    LOGGER.warning("User aborted open command.")
                elif choice == "all":
                    LOGGER.debug("User selected all urls.")
                    for url in url_list:
                        self._open_url(url)
                elif choice in things_to_open.keys():
                    LOGGER.debug("User selected the %s set of urls.", choice)
                    for url in things_to_open[choice]:
                        self._open_url(url)
                elif choice.isdigit():
                    LOGGER.debug("User selected url %s", choice)
                    self._open_url(url_list[int(choice) - 1])

        Event.PostOpenCommand.fire(self)

    @staticmethod
    def _open_url(url: ParseResult) -> None:
        """Opens a URL.

        Args:
            url: the URL to be opened.
        """
        opener = config.commands.open.command
        try:
            url_str: str = url.geturl() if url.scheme else str(RelPath(url.geturl()).path)
            LOGGER.debug('Opening "%s" with %s.', url_str, opener)
            with open(os.devnull, "w", encoding="utf-8") as devnull:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=ResourceWarning)
                    subprocess.Popen(  # pylint: disable=consider-using-with
                        [opener, url_str],
                        stdout=devnull,
                        stderr=devnull,
                        stdin=devnull,
                        close_fds=True,
                    )
        except FileNotFoundError as err:
            LOGGER.error(err)

    @staticmethod
    def _wrap_prompt_process_response(
        func: Callable[[PromptBase[PromptType], str], PromptType]
    ) -> Callable[[PromptBase[PromptType], str], PromptType]:
        """A method to wrap a `PromptBase.process_response` method.

        This method wraps a `PromptBase.process_response` method in order to handle a user's request
        for additional help.

        Args:
            func: the `PromptBase.process_response` method to be wrapped.

        Returns:
            The wrapped `PromptBase.process_response` method.
        """

        @override
        @wraps(func)
        def process_response(prompt: PromptBase[PromptType], value: str) -> PromptType:
            return_value: PromptType = func(prompt, value)

            if return_value == "help":
                LOGGER.debug("User requested help.")
                raise InvalidResponse(
                    "[yellow]Multiple targets were found. You may select the following:\n"
                    "  1. an individual URL number\n"
                    "  2. a target type (provided in '[...]')\n"
                    "  3. 'all'\n"
                    "  4. or 'cancel' to abort the command"
                )

            return return_value

        return process_response



r"""coBib's List command.

This command simply lists the entries in the database:
```
cobib list
```
By default, it lists them in the order in which they appear in the database. However, the order as
well as selection of entries to be listed can be manipulated via additional arguments as explained
below.

### Basic Options

It provides some very basic output manipulation options.

You can reverse the output:
```
cobib list --reverse
```
You can sort the list according to a database field:
```
cobib list --sort year
```
(In the TUI, this is available via the `s` key, by default.)


### Filters

However, the arguably most useful feature of this command are the *filters*.

These are a set of keyword arguments which are registered at runtime depending on the fields which
appear in your database.
To give an example, you can find the following filters in the output of `cobib list --help`:
```
++author AUTHOR
--author AUTHOR
++year YEAR
--year YEAR
```
As you can see, each field is used twice: once with a `++` and once with a `--` prefix.
These allow you to specify positive and negative matches, respectively.
For example:
```
cobib list ++year 2020
```
will list *only* entries whose `year` contains `2020`.
On the contrary:
```
cobib list --year 2020
```
will print all those entries whose `year` does *not* contain `2020`.

You can combine multiple filters to narrow your selection down further.
For example:
```
cobib list ++year 2020 ++author Rossmannek
```
will list entries whose `year` contains `2020` *and* whose `author` field contains `Rossmannek`.

Note, that by default you can use the `f` key in the TUI to add filters to the displayed list of
entries.

There are some aspects to take note of here:
1. By default, multiple filters are combined with logical `AND`s. You can specify `--or` to
   overwrite this to logical `OR`s. This will apply to all filters of the specified command.
2. All entries are treated as `str`. Thus, `++year 20` will match anything *containing* `20`.

As of version v3.2.0, the filter arguments are evaluated as regex patterns allowing you to do things
like the following:
```
cobib list ++label "\D+_\d+"
```
This will list all entries whose labels are formatted as `"<non-digit characters>_<digits>"`.

As of version v4.0.0, you can make the filter matching case-**in**sensitive via the
`cobib.config.config.ListCommandConfig.ignore_case` setting which defaults to being `False`.
Besides this setting, you can always overwrite its value on the command line with the
`--ignore-case` (`-i` for short) and `--no-ignore-case` (`-I` for short; since v4.1.0) options.
Providing these options takes precedence over your configuration value.
Thus, the following will *always* still match entries where `Rossmannek` is contained in the author
field:
```
cobib list --ignore-case ++author rossmannek
```
And the following will *always* be sensitive to case:
```
cobib list --no-ignore-case ++author rossmannek
```

.. note::
   For more information on the filtering mechanisms see also `cobib.database.Entry.matches`.
"""

from __future__ import annotations

import argparse
import logging
from collections import defaultdict
from copy import copy
from typing import Any, Dict, List, Optional, Set, Tuple, Type

from rich.console import Console, ConsoleRenderable
from rich.prompt import PromptBase, PromptType
from rich.table import Table
from rich.text import Text
from textual.app import App
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.ui.components import ListView

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class ListCommand(Command):
    """The List Command.

    This command can parse the following arguments:

        * `-s`, `--sort`: you can specify an arbitrary `cobib.database.Entry.data` field name which
          should be used for sorting the listed entries. This will automatically include a column
          for this field in the output table.
        * `-r`, `--reverse`: if specified, the entries will be listed in reverse order. This is
          especially useful in the TUI (where it is enabled by default) because it puts the last
          added entries at the top of the window. When using the command-line interface it is
          disabled by default, because this puts the last added entries at the bottom, just above
          the new command-line prompt.
        * `-i`, `--ignore-case`: if specified, the entry matching will be case-**in**sensitive. This
          overwrites the `cobib.config.config.ListCommandConfig.ignore_case` setting.
        * `-I`, `--no-ignore-case`: if specified, the entry matching will be case-sensitive. This
          overwrites the `cobib.config.config.ListCommandConfig.ignore_case` setting.
        * `-x`, `--or`: if specified, multiple filters will be combined with logical OR rather than
          the default logical AND.
        * in addition to the options above, [Filter keyword arguments](#filters) are registered at
          runtime based on the fields available in the database. Please refer that section or the
          output of `cobib list --help` for more information.
    """

    name = "list"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.entries: List[Entry] = []
        """A list of entries, filtered and sorted according to the provided command arguments."""

        self.columns: List[str] = []
        """A list of (key) columns to be included when rendering the results."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(
            prog="list", description="List subcommand parser.", prefix_chars="+-"
        )
        parser.add_argument("-s", "--sort", help="specify column along which to sort the list")
        parser.add_argument(
            "-r", "--reverse", action="store_true", help="reverses the listing order"
        )
        ignore_case_group = parser.add_mutually_exclusive_group()
        ignore_case_group.add_argument(
            "-i",
            "--ignore-case",
            action="store_true",
            default=None,
            help="ignore case for entry matching",
        )
        ignore_case_group.add_argument(
            "-I",
            "--no-ignore-case",
            dest="ignore_case",
            action="store_false",
            default=None,
            help="do NOT ignore case for entry matching",
        )
        parser.add_argument(
            "-x",
            "--or",
            dest="OR",
            action="store_true",
            help="concatenate filters with OR instead of AND",
        )
        unique_keys: Set[str] = {"label"}
        LOGGER.debug("Gathering possible filter arguments.")
        for entry in Database().values():
            unique_keys.update(entry.data.keys())
        for key in sorted(unique_keys):
            parser.add_argument(
                "++" + key, type=str, action="append", help="include elements with matching " + key
            )
            parser.add_argument(
                "--" + key, type=str, action="append", help="exclude elements with matching " + key
            )

        cls.argparser = parser

    @override
    @classmethod
    def _parse_args(cls, args: tuple[str, ...]) -> argparse.Namespace:
        args = tuple(arg for arg in args if arg != "--")

        return super()._parse_args(args)

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting List command.")

        Event.PreListCommand.fire(self)

        filtered_entries, filtered_keys = self.filter_entries()

        self.entries = self._sort_entries(filtered_entries, self.largs.sort, self.largs.reverse)

        # construct list of columns to be displayed
        self.columns = copy(config.commands.list_.default_columns)
        LOGGER.debug("Listing the default columns: %s", str(self.columns))
        # display the column along which was sorted
        if self.largs.sort and self.largs.sort not in self.columns:
            LOGGER.debug("Appending the column which is sorted by: %s", str(self.largs.sort))
            self.columns.append(self.largs.sort)
        # also display the keys which were used to filter
        LOGGER.debug("Extendings the columns which are filtered by: %s", str(filtered_keys))
        self.columns.extend(col for col in filtered_keys if col not in self.columns)

        Event.PostListCommand.fire(self)

    def filter_entries(self) -> Tuple[List[Entry], Set[str]]:
        """The filtering method.

        This method implements the actual filtering routine. Based on the arguments provided to this
        command, this method will iterate the database and return those entries which match the
        specified filter.

        Returns:
            A pair indicating the matching entries. The first object is the list of matching entries
            (which is also exposed via `entries`). The second object is the set of keys which were
            filtered on. This can be used (for example) to include these keys during the result
            rendering.
        """
        LOGGER.debug("Constructing filter.")

        filtered_keys: Set[str] = set()
        _filter: Dict[Tuple[str, bool], List[Any]] = defaultdict(list)

        for key, val in self.largs.__dict__.items():
            if key in ["OR", "sort", "reverse", "ignore_case"] or val is None:
                # ignore special arguments
                continue

            # track the keys being filtered to display these columns later
            filtered_keys.add(key)

            if not isinstance(val, list):
                val = [val]
            # iterate values to be filtered by
            for i in val:
                for idx, obj in enumerate(self.args):
                    if i == obj:
                        # once we find the current value in the CLI argument list we can determine
                        # whether this filter is INclusive (`++`) or EXclusive (`--`)
                        index: Tuple[str, bool] = (key, self.args[idx - 1][0] == "+")
                        _filter[index].append(i)
                        break

        LOGGER.debug("Final filter configuration: %s", dict(_filter))

        if self.largs.OR:
            LOGGER.debug("Filters are combined with logical ORs!")

        ignore_case = config.commands.list_.ignore_case
        if self.largs.ignore_case is not None:
            ignore_case = self.largs.ignore_case
        LOGGER.debug(
            "The entry matching will be performed case %ssensitive", "in" if ignore_case else ""
        )

        for key, entry in Database().items():
            if entry.matches(_filter, self.largs.OR, ignore_case):
                LOGGER.debug('Entry "%s" matches the filter.', key)
                self.entries.append(entry)

        return self.entries, filtered_keys

    @staticmethod
    def _sort_entries(
        entries: List[Entry], sort: Optional[str] = None, reverse: bool = False
    ) -> List[Entry]:
        """The sorting method.

        This method sorts the provided entries according to the requested key and order.

        Args:
            entries: the list of entries to be sorted.
            sort: the optional key by which to sort.
            reverse: whether or not to sort in reverse order.

        Returns:
            The sorted list of entries.
        """
        if reverse:
            LOGGER.debug("Reversing the entry order.")

        if sort is None:
            if reverse:
                return entries[::-1]
            return entries

        LOGGER.debug("Sorting entries by key '%s'.", sort)

        sorted_entries: List[Entry] = sorted(
            entries, reverse=reverse, key=lambda entry: entry.stringify().get(str(sort), "")
        )

        return sorted_entries

    @override
    def render_porcelain(self) -> List[str]:
        output: List[str] = []

        output.append("::".join(self.columns))

        for entry in self.entries:
            stringified: Dict[str, str] = entry.stringify()

            output.append("::".join(stringified.get(col, "") for col in self.columns))

        return output

    @override
    def render_rich(self) -> ConsoleRenderable:
        rich_table = Table()

        for col in self.columns:
            rich_table.add_column(col)

        for entry in self.entries:
            stringified: Dict[str, str] = entry.stringify()

            rich_table.add_row(*(stringified.get(col, "") for col in self.columns))

        return rich_table

    @override
    def render_textual(self) -> ListView:
        textual_table = ListView()

        for col in self.columns:
            textual_table.add_column(col, width=None)

        for entry in self.entries:
            stringified: Dict[str, str] = entry.stringify()

            textual_table.add_row(*(Text(stringified.get(col, "")) for col in self.columns))

        return textual_table



"""coBib's Command interface."""

from __future__ import annotations

import argparse
import json
import logging
import os
import shlex
import sys
from abc import ABC, abstractmethod
from typing import List, Optional, Type

from rich.console import Console, ConsoleRenderable
from rich.prompt import Prompt, PromptBase, PromptType
from textual.app import App
from textual.widget import Widget

from cobib.config import Event, config
from cobib.ui.components import ArgumentParser as ArgumentParser
from cobib.utils.rel_path import RelPath

LOGGER = logging.getLogger(__name__)


class Command(ABC):
    """The Command interface.

    This interface should be implemented by all concrete command implementations.
    """

    name = "base"
    """The commands `name` is used to extract the available commands for the command-line interface.
    """

    argparser: ArgumentParser
    """Every command has its own `argparse.ArgumentParser` which is used to parse the arguments
    provided to the command. This is done no matter how the command is executed, whether
    programmatically via Python, from the command-line or any other UI."""

    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        """Initializes a command instance.

        Args:
            *args: the sequence of additional command arguments. These will be passed on to the
                `argparser` of this command for further parsing.
            console: a command may be in need of printing something for the user during its
                execution (and before its final result rendering). If so, it should use the `print`
                method of this console object.
            prompt: a command may be in need of prompting the user for an input. If so, it will use
                this prompt kind. It is important that this is respected, as different UIs may
                inject specific prompt classes to implement different runtime behavior.
        """
        self.args: tuple[str, ...] = args
        """The raw provided command arguments."""

        self.largs: argparse.Namespace = self.__class__._parse_args(args)
        """The parsed (local) arguments."""

        self.console: Console | App[None] = console if console is not None else Console()
        """The object via which to print output to the user during runtime execution."""

        self.prompt: Type[PromptBase[PromptType]] = prompt if prompt is not None else Prompt
        """The object via which to prompt the user for input during runtime execution."""

    @classmethod
    @abstractmethod
    def init_argparser(cls) -> None:
        """Initializes this command's `argparse.ArgumentParser`.

        This method needs to be overwritten by every subclass and handles the registration of all
        available command arguments.
        """

    @classmethod
    def _get_argparser(cls) -> ArgumentParser:
        """Returns this command's `argparse.ArgumentParser`.

        The reason for having this method is to handle the parser initialization such that it only
        needs to be done once.

        Returns:
            This command's initialized `argparser` object.
        """
        if hasattr(cls, "argparser"):
            return cls.argparser

        cls.init_argparser()
        return cls.argparser

    @classmethod
    def _parse_args(cls, args: tuple[str, ...]) -> argparse.Namespace:
        """Parses the provided command arguments.

        Args:
            args: the sequence of additional command arguments provided to the command upon
                initialization.

        Returns:
            The parsed arguments namespace.
        """
        try:
            largs = cls._get_argparser().parse_args(args)
        except argparse.ArgumentError as exc:
            LOGGER.error(exc.message)
            sys.exit(1)

        return largs

    @abstractmethod
    def execute(self) -> None:
        """Actually executes the command.

        This means, all of the command-specific logic and action needs to be implemented by this
        method.

        .. note::
           This method is **not** in charge of presenting the final result to the user. Refer to the
           various `render_*` methods, instead.

        As a consequence, resulting data should be stored on the command instance (which also has
        the benefit of exposing this data to the various `Post*Command` `cobib.config.event.Event`
        hooks.

        This function may *not* raise any errors!
        This also means it should take care of catching all potential errors triggered by internally
        used methods.
        Such encountered errors should be logged and the method should return gracefully.
        """

    def render_porcelain(self) -> List[str]:
        """Renders the command results in "porcelain" mode.

        This method is called when the `--porcelain` argument has been provided.
        The idea is to provide an output mode which is easily parse-able by another program or
        function.

        Returns:
            A list of strings where each entry should be considered one line of output.
        """
        return []

    def render_rich(self) -> Optional[ConsoleRenderable]:
        """Renders the command results as a `rich` object.

        This method is called when a command is run via the command-line interface.

        Returns:
            An optional `ConsoleRenderable` to be presented to the user.
        """
        return None

    def render_textual(self) -> Optional[Widget]:
        """Renders the command results as a `textual` widget.

        This method is called when a command is run via the terminal user interface.
        It is the responsibility of the TUI to deal with the returned widget.

        Returns:
            An optional `Widget` to be rendered in the TUI.
        """
        return None

    def git(self, force: bool = False) -> None:
        """Generates a git commit to track the commands changes.

        This function only has an effect when `cobib.config.config.DatabaseConfig.git` is enabled
        *and* the database has been initialized correctly with `cobib.commands.init`.
        Otherwise, a warning will be printed and no commit will be generated.
        Nonetheless, the changes applied by the commit will have taken effect in the database.

        This method uses the parsed arguments (`largs`) to include command execution information in
        the generated commit message.

        Args:
            force: whether to ignore the configuration setting. This option is mainly used by the
                `cobib.commands.init.InitCommand`.
        """
        git_tracked = config.database.git
        if not git_tracked and not force:
            return

        file = RelPath(config.database.file).path
        root = file.parent

        if not (root / ".git").exists():
            if git_tracked:
                msg = (
                    "You have configured coBib to track your database with git."
                    "\nPlease run `cobib init --git`, to initialize this tracking."
                )
                LOGGER.warning(msg)
                return

        args = vars(self.largs)
        msg = f"Auto-commit: {self.name.title()}Command"
        if args:
            msg += "\n\n"
            msg += json.dumps(args, indent=2, default=str)

        msg = Event.PreGitCommit.fire(msg, args) or msg

        commands = [
            f"cd {root}",
            f"git add -- {file}",
            f"git commit --no-gpg-sign --quiet --message {shlex.quote(msg)}",
        ]
        LOGGER.debug("Auto-commit to git from %s command.", self.name)
        os.system("; ".join(commands))

        Event.PostGitCommit.fire(root, file)



"""Tests for coBib's UndoCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import contextlib
import logging
import os
import subprocess
from io import StringIO
from shutil import rmtree
from typing import TYPE_CHECKING, Any, Type

import pytest
from typing_extensions import override

from cobib.commands import AddCommand, UndoCommand
from cobib.config import Event, config
from cobib.database import Database

from .. import get_resource
from .command_test import CommandTest

EXAMPLE_MULTI_FILE_ENTRY_BIB = get_resource("example_multi_file_entry.bib", "commands")

if TYPE_CHECKING:
    import cobib.commands


class TestUndoCommand(CommandTest):
    """Tests for coBib's UndoCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return UndoCommand

    def _assert(self) -> None:
        """Common assertion utility method."""
        assert Database().get("example_multi_file_entry", None) is None

        # get last commit message
        with subprocess.Popen(
            ["git", "-C", self.COBIB_TEST_DIR, "show", "--format=format:%B", "--no-patch", "HEAD"],
            stdout=subprocess.PIPE,
        ) as proc:
            message, _ = proc.communicate()
            # decode it
            split_message = message.decode("utf-8").split("\n")
            # assert subject line
            assert "Undo" in split_message[0]

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup", "expected_exit"],
        [
            [{"git": False}, False],
            [{"git": True}, False],
            [{"git": True}, True],
        ],
        indirect=["setup"],
    )
    async def test_command(
        self, setup: Any, expected_exit: bool, caplog: pytest.LogCaptureFixture
    ) -> None:
        # pylint: disable=invalid-overridden-method
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            expected_exit: whether to expect an early exit.
            caplog: the built-in pytest fixture.
        """
        git = setup.get("git", False)

        if not git:
            UndoCommand().execute()
            for source, level, message in caplog.record_tuples:
                if ("cobib.commands.undo", logging.ERROR) == (
                    source,
                    level,
                ) and "git-tracking" in message:
                    break
            else:
                pytest.fail("No Error logged from UndoCommand.")
        elif expected_exit:
            # Regression test related to #65
            with pytest.raises(SystemExit):
                UndoCommand().execute()
            for source, level, message in caplog.record_tuples:
                if ("cobib.commands.undo", logging.WARNING) == (
                    source,
                    level,
                ) and "Could not find a commit to undo." in message:
                    break
            else:
                pytest.fail("No Error logged from UndoCommand.")
        else:
            await AddCommand("-b", EXAMPLE_MULTI_FILE_ENTRY_BIB).execute()
            UndoCommand().execute()

            self._assert()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    async def test_skipping_undone_commits(
        self, setup: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test skipping already undone commits.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        await AddCommand("-b", EXAMPLE_MULTI_FILE_ENTRY_BIB).execute()
        await AddCommand("-b", get_resource("example_entry.bib")).execute()
        UndoCommand().execute()
        caplog.clear()

        UndoCommand().execute()
        self._assert()
        assert "Storing undone commit" in caplog.record_tuples[4][2]
        assert "Skipping" in caplog.record_tuples[6][2]

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    def test_warn_insufficient_setup(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning in case of insufficient setup.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        rmtree(self.COBIB_TEST_DIR_GIT)
        UndoCommand().execute()
        for source, level, message in caplog.record_tuples:
            if ("cobib.commands.undo", logging.ERROR) == (
                source,
                level,
            ) and "configured, but not initialized" in message:
                break
        else:
            pytest.fail("No Error logged from UndoCommand.")

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    # other variants are already covered by test_command
    async def test_cmdline(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            caplog: the built-in pytest fixture.
        """
        await AddCommand("-b", EXAMPLE_MULTI_FILE_ENTRY_BIB).execute()
        await self.run_module(monkeypatch, "main", ["cobib", "undo"])

        self._assert()

    # manually overwrite this test because we must enable git integration
    def test_handle_argument_error(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test handling of ArgumentError.

        Args:
            caplog: the built-in pytest fixture.
        """
        # use temporary config
        config.database.file = self.COBIB_TEST_DIR / "database.yaml"
        config.database.git = True

        # initialize git-tracking
        self.COBIB_TEST_DIR.mkdir(parents=True, exist_ok=True)
        open(  # pylint: disable=consider-using-with
            config.database.file, "w", encoding="utf-8"
        ).close()
        os.system("git init " + str(self.COBIB_TEST_DIR))

        try:
            super().test_handle_argument_error(caplog)
        finally:
            # clean up file system
            rmtree(self.COBIB_TEST_DIR_GIT)
            # clean up config
            config.defaults()

    @pytest.mark.parametrize("setup", [{"git": True}], indirect=["setup"])
    def test_event_pre_undo_command(self, setup: Any) -> None:
        """Tests the PreUndoCommand event."""

        @Event.PreUndoCommand.subscribe
        def hook(command: UndoCommand) -> None:
            print("Hello world!")

        assert Event.PreUndoCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as out:
            with pytest.raises(SystemExit):
                UndoCommand().execute()

            assert out.getvalue() == "Hello world!\n"

    @pytest.mark.asyncio
    @pytest.mark.parametrize("setup", [{"git": True}], indirect=["setup"])
    async def test_event_post_undo_command(self, setup: Any) -> None:
        """Tests the PostUndoCommand event."""

        @Event.PostUndoCommand.subscribe
        def hook(command: UndoCommand) -> None:
            print(command.root)

        assert Event.PostUndoCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as out:
            await AddCommand("-b", EXAMPLE_MULTI_FILE_ENTRY_BIB).execute()
            UndoCommand().execute()

            self._assert()

            assert out.getvalue() == f"{self.COBIB_TEST_DIR}\n"



"""coBib's Entry class."""

from __future__ import annotations

import logging
import re
import subprocess
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from pylatexenc.latexencode import UnicodeToLatexEncoder

from cobib.config import config
from cobib.utils.rel_path import RelPath

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    import cobib.parsers


class Entry:
    """coBib's bibliographic entry.

    coBib's `Database` stores the bibliographic information in entries which are instances of this
    class. This only contains a `label` which is a string used as a key to associate this entry as
    well as a free-form `data` dictionary, which can contain arbitrary key-value pairs.
    One field will *always* be present in the `data` dictionary:
    * `ENTRYTYPE`: which specifies the BibLaTex type of the entry.

    Only through the context imposed by BibLaTex will the other `data` fields be interpreted.

    However, some keys are exposed as properties in a special format to provide easy access to these
    meta-data fields not typically used by BibLaTex itself.

    Besides being a data container, this class also provides some utilities for data manipulation
    and querying.
    """

    def __init__(self, label: str, data: Dict[str, Any]) -> None:
        """Initializes a new Entry.

        Args:
            label: the label associated with this entry in the `Database`.
            data: the actual bibliographic data stored as a dictionary mapping free-form field names
                (`str`) to any other data. Some fields are exposed as properties of this class for
                convenience.
        """
        LOGGER.debug("Initializing entry: %s", label)

        self._label: str = str(label)

        self.data: Dict[str, Any] = {}
        """The actual bibliographic data."""

        for key, value in data.items():
            if hasattr(self, key):
                setattr(self, key, value)
            elif isinstance(value, str) and value.isnumeric():
                LOGGER.info(
                    "Converting field '%s' of entry '%s' to integer: %s.",
                    key,
                    label,
                    value,
                    extra={"entry": label, "field": key},
                )
                self.data[key] = int(value)
            else:
                self.data[key] = value

        if "ID" in self.data:
            self.data.pop("ID")
            LOGGER.info(
                "The field '%s' of entry '%s' is no longer required. It will be inferred from the "
                "entry label.",
                "ID",
                label,
                extra={"entry": label, "field": "ID"},
            )

    def __eq__(self, other: object) -> bool:
        """Checks equality of two entries."""
        if not isinstance(other, Entry):
            return False
        return self.label == other.label and self.data == other.data

    @property
    def label(self) -> str:
        """The `Database` label of this entry."""
        return self._label

    @label.setter
    def label(self, label: str) -> None:
        """Sets the `Database` label of this entry.

        Args:
            label: the label of this entry.
        """
        LOGGER.debug("Changing the label '%s' to '%s'.", self.label, label)
        self._label = str(label)

    @property
    def tags(self) -> List[str]:
        """The tags of this entry."""
        return self.data.get("tags", [])

    @tags.setter
    def tags(self, tags: Union[str, List[str]]) -> None:
        """Sets the tags of this entry.

        Args:
            tags: a single or list of tags.
        """
        if isinstance(tags, list):
            self.data["tags"] = tags
        else:
            self.data["tags"] = tags.split(config.database.stringify.list_separator.tags)
            if len(self.data["tags"]) > 1:
                LOGGER.info(
                    "Converted the field '%s' of entry '%s' to a list. You can consider storing it "
                    "as such directly.",
                    "tags",
                    self.label,
                    extra={"entry": self.label, "field": "tags"},
                )
        LOGGER.debug("Adding the tags '%s' to '%s'.", self.data["tags"], self.label)

    @property
    def file(self) -> List[str]:
        # noqa: D402 (we skip this error because file(s) raises a false negative)
        """The associated file(s) of this entry.

        The setter of this property will convert the strings to paths relative to the user's home
        directory. Internally, this field will always be stored as a list.
        """
        return self.data.get("file", [])

    @file.setter
    def file(self, file: Union[str, List[str]]) -> None:
        # noqa: D402 (we skip this error because file(s) raises a false negative)
        """Sets the associated file(s) of this entry.

        Args:
            file: can be either a single path (`str`) or a list thereof. In either case, the strings
                will be converted to paths relative to the user's home directory. Internally, this
                field will always be stored as a list.
        """
        if isinstance(file, list):
            paths = [RelPath(f) for f in file]
        else:
            paths = [RelPath(f) for f in file.split(config.database.stringify.list_separator.file)]
            if len(paths) > 1:
                LOGGER.info(
                    "Converted the field '%s' of entry '%s' to a list. You can consider storing it "
                    "as such directly.",
                    "file",
                    self.label,
                    extra={"entry": self.label, "field": "file"},
                )
        self.data["file"] = [str(p) for p in paths]
        LOGGER.debug("Adding '%s' as the file to '%s'.", self.data["file"], self.label)

    @property
    def url(self) -> List[str]:
        """The associated URL(s) of this entry."""
        return self.data.get("url", [])

    @url.setter
    def url(self, url: Union[str, List[str]]) -> None:
        """Sets the associated URL(s) of this entry.

        Args:
            url: can be either a single URL (`str`) or a list thereof. Internally, this field will
            always be stored as a list.
        """
        if isinstance(url, list):
            self.data["url"] = url
        else:
            self.data["url"] = url.split(config.database.stringify.list_separator.url)
            if len(self.data["url"]) > 1:
                LOGGER.info(
                    "Converted the field '%s' of entry '%s' to a list. You can consider storing it "
                    "as such directly.",
                    "url",
                    self.label,
                    extra={"entry": self.label, "field": "url"},
                )
        LOGGER.debug("Adding '%s' as the url to '%s'.", self.data["url"], self.label)

    @property
    def month(self) -> str:
        """Returns the month."""
        return self.data.get("month", None)

    @month.setter
    def month(self, month: Union[int, str]) -> None:
        """Sets the month."""
        months = [
            "jan",
            "feb",
            "mar",
            "apr",
            "may",
            "jun",
            "jul",
            "aug",
            "sep",
            "oct",
            "nov",
            "dec",
        ]
        if month in months:
            self.data["month"] = month
        else:
            if isinstance(month, int):
                self.data["month"] = months[month - 1]
            elif isinstance(month, str):
                if month.isnumeric():
                    self.data["month"] = months[int(month) - 1]
                else:
                    self.data["month"] = month.lower()[:3]
            LOGGER.info(
                "Converting field '%s' of entry '%s' from '%s' to '%s'.",
                "month",
                self.label,
                month,
                self.data["month"],
                extra={"entry": self.label, "field": "month"},
            )

    def stringify(self) -> Dict[str, str]:
        """Returns an identical entry to self but with all fields converted to strings.

        Returns:
            An `Entry` with purely string fields.
        """
        data = {}
        data["label"] = self.label
        for field, value in self.data.items():
            if isinstance(value, list) and hasattr(config.database.stringify.list_separator, field):
                data[field] = getattr(config.database.stringify.list_separator, field).join(value)
            else:
                data[field] = str(value)
        return data

    def escape_special_chars(self, suppress_warnings: bool = True) -> None:
        """Escapes special characters in the bibliographic data.

        Special characters should be escaped to ensure proper rendering in LaTeX documents. This
        function leverages the existing implementation of the `pylatexenc` module to do said
        conversion. The only fields exempted from the conversion are the `file` and `url` fields of
        the `Entry.data` dictionary.

        Args:
            suppress_warnings: if True, warnings generated by the `pylatexenc` modules will be
                suppressed. This argument will be overwritten if the logging level is set to
                `logging.DEBUG`.
        """
        enc = UnicodeToLatexEncoder(
            non_ascii_only=True,
            replacement_latex_protection="braces-all",
            unknown_char_policy="keep",
            unknown_char_warning=not suppress_warnings or LOGGER.isEnabledFor(logging.DEBUG),
        )
        for key, value in self.data.items():
            if key in ("file", "url"):
                # do NOT these fields and keep any special characters
                self.data[key] = value
                continue
            if isinstance(value, str):
                self.data[key] = enc.unicode_to_latex(value)

    def save(self, parser: Optional[cobib.parsers.base_parser.Parser] = None) -> str:
        """Saves an entry using the parsers `dump` method.

        This method is mainly used by the `Database.save` method and takes care of some final
        conversions depending on the user's configuration. Applying such modifications (like e.g.
        special character escaping) only before saving ensures a consistent state of the database
        while also providing a fast startup because these conversions are prevented at that time.

        Args:
            parser: the parser instance to use for dumping. If set to `None` it will default to a
                `cobib.parsers.YAMLParser`. Supplying a ready instance can improve efficiency
                significantly while saving many entries after one another.

        Returns:
            The string-representation of this entry as produced by the provided parser.
        """
        self.escape_special_chars(config.database.format.suppress_latex_warnings)
        if parser is None:
            # pylint: disable=import-outside-toplevel,cyclic-import
            from cobib.parsers.yaml import YAMLParser

            parser = YAMLParser()
        return parser.dump(self) or ""  # `dump` may return `None`

    def matches(
        self, filter_: Dict[Tuple[str, bool], List[str]], or_: bool, ignore_case: bool = False
    ) -> bool:
        """Check whether this entry matches the supplied filter.

        coBib provides an extensive filtering implementation. The filter is specified in the form
        of a dictionary whose keys consist of pairs of `(str, bool)` entries where the string
        indicates the field to match against and the boolean whether a positive (`true`) or negative
        (`false`) match is required. The values of the dictionary must be a `List[str]`. This means
        that field types are always compared on a string-basis (this is a limitation of the targeted
        command-line interface). However, as of v3.2.0 these strings are interpreted as regex
        patterns providing the most power within this framework.

        Some examples:

        | `filter_`                             | `or_`    | Meaning                               |
        | ------------------------------------- | -------- | ------------------------------------- |
        | `{('year', True): ['2020']}`          | *either* | `year` contains 2020                  |
        | `{('year', False): ['2020']}`         | *either* | `year` does not contain 2020          |
        | `{('year', True): ['2020', '2021']}`  | True     | `year` contains either 2020 or 2021   |
        | `{('year', True): ['2020', '2021']}`  | False    | cannot match anything                 |
        | `{('year', False): ['2020', '2021']}` | False    | `year` contains neither 2020 nor 2021 |

        Args:
            filter_: dictionary describing the filter as explained above.
            or_ : boolean indicating whether logical OR (`true`) or AND (`false`) is used to combine
                multiple filter items.
            ignore_case: if True, the matching will be case-*in*sensitive.

        Returns:
            Boolean indicating whether this entry matches the filter.
        """
        LOGGER.debug("Checking whether entry %s matches.", self.label)
        re_flags = re.IGNORECASE if ignore_case else 0
        match_list = []
        stringified_data = self.stringify()
        for key, values in filter_.items():
            if key[0] not in stringified_data:
                match_list.append(not key[1])
                continue
            for val in values:
                if re.search(rf"{val}", stringified_data[key[0]], flags=re_flags):
                    match_list.append(key[1])
                else:
                    match_list.append(not key[1])
        if or_:
            return any(m for m in match_list)
        return all(m for m in match_list)

    def search(
        self, query: List[str], context: int = 1, ignore_case: bool = False
    ) -> List[List[str]]:
        """Search entry contents for the query strings.

        The entry will *always* be converted to a searchable string using the
        `cobib.parsers.BibtexParser.dump` method. This text will then be search for each item in
        `query` and will interpret these as regex patterns.
        If a `file` is associated with this entry, the search will try its best to recursively query
        its contents, too. However, the success of this depends highly on the configured search
        tool, `cobib.config.config.SearchCommandConfig.grep`.

        Args:
            query: the list of regex patterns to search for.
            context: the number of context lines to provide for each match. This behaves similarly
                to the *Context Line Control* available for the UNIX `grep` command (`--context`).
            ignore_case: if True, the search will be case-*in*sensitive.

        Returns:
            A list of lists containing the context for each match associated with this entry.
        """
        LOGGER.debug("Searching entry %s.", self.label)
        matches: List[List[str]] = []
        # pylint: disable=import-outside-toplevel,cyclic-import
        from cobib.parsers.bibtex import BibtexParser

        bibtex = BibtexParser().dump(self).split("\n")
        re_flags = re.IGNORECASE if ignore_case else 0
        for query_str in query:
            re_compiled = re.compile(rf"{query_str}", flags=re_flags)
            for idx, line in enumerate(bibtex):
                if re_compiled.search(line):
                    # add new match
                    matches.append([])
                    # upper context; (we iterate in reverse in order to ensure that we abort on the
                    # first previous occurrence of the query pattern)
                    for string in reversed(bibtex[max(idx - context, 0) : min(idx, len(bibtex))]):
                        if re_compiled.search(string):
                            break
                        matches[-1].insert(0, string)
                    # matching line itself
                    matches[-1].append(line)
                    # lower context
                    for string in bibtex[max(idx + 1, 0) : min(idx + context + 1, len(bibtex))]:
                        if re_compiled.search(string):
                            break
                        matches[-1].append(string)

            for file_ in self.file:
                grep_prog = config.commands.search.grep
                path = RelPath(file_).path
                if not path.exists():
                    LOGGER.warning(
                        "The associated file %s of entry %s does not exist!", file_, self.label
                    )
                    continue

                LOGGER.debug("Searching associated file %s with %s", file_, grep_prog)
                with subprocess.Popen(
                    [
                        grep_prog,
                        *config.commands.search.grep_args,
                        f"-C{context}",
                        query_str,
                        RelPath(file_).path,
                    ],
                    stdout=subprocess.PIPE,
                ) as grep:
                    if grep.stdout is None:
                        continue
                    stdout = grep.stdout
                    # extract results
                    results = stdout.read().decode().split("\n--\n")
                for match in results:
                    if match:
                        matches.append([line.strip() for line in match.split("\n") if line.strip()])

        return matches



#!/usr/bin/env python3
# pylint: disable=missing-docstring
import os
import setuptools

HERE = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(HERE, 'requirements.txt'), 'r') as f:
    REQUIREMENTS = f.read().strip().split('\n')

setuptools.setup(install_requires=REQUIREMENTS)



"""Tests for coBib's ImportCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type

import pytest
from typing_extensions import override

from cobib.commands import ImportCommand
from cobib.config import Event

from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestImportCommand(CommandTest):
    """Tests for coBib's ImportCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return ImportCommand

    @pytest.mark.asyncio
    async def test_command(self) -> None:  # type: ignore[override]
        # pylint: disable=invalid-overridden-method
        """Test the command itself."""
        with pytest.raises(SystemExit):
            await ImportCommand("-h").execute()

    @pytest.mark.asyncio
    async def test_cmdline(self, setup: Any, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
        """
        with pytest.raises(SystemExit):
            await self.run_module(monkeypatch, "main", ["cobib", "import", "-h"])

    @pytest.mark.asyncio
    async def test_event_pre_import_command(self, setup: Any) -> None:
        """Tests the PreImportCommand event."""

        @Event.PreImportCommand.subscribe
        def hook(command: ImportCommand) -> None:
            command.largs.zotero = False

        assert Event.PreImportCommand.validate()

        await ImportCommand("--zotero").execute()

    @pytest.mark.asyncio
    async def test_event_post_import_command(self, setup: Any) -> None:
        """Tests the PostImportCommand event."""

        @Event.PostImportCommand.subscribe
        def hook(command: ImportCommand) -> None:
            pass

        assert Event.PostImportCommand.validate()

        @Event.PreImportCommand.subscribe
        def aux_hook(command: ImportCommand) -> None:
            command.largs.zotero = False

        await ImportCommand("--zotero").execute()



"""Tests for coBib's YAMLParser."""
# pylint: disable=unused-argument

from typing import Dict, Optional, cast

import pytest

from cobib.config import Event, config
from cobib.database import Entry
from cobib.parsers import YAMLParser

from .parser_test import ParserTest


class TestYAMLParser(ParserTest):
    """Tests for coBib's YAMLParser."""

    def test_to_yaml(self) -> None:
        """Test to yaml conversion."""
        entry = Entry("Cao_2019", self.EXAMPLE_ENTRY_DICT)
        yaml_str = YAMLParser().dump(entry)
        with open(self.EXAMPLE_YAML_FILE, "r", encoding="utf-8") as file:
            assert yaml_str == file.read()

    @pytest.mark.parametrize("use_c_lib_yaml", [False, True])
    def test_from_yaml_file(self, use_c_lib_yaml: bool) -> None:
        """Test parsing a yaml file.

        Args:
            use_c_lib_yaml: the configuration setting.
        """
        try:
            config.parsers.yaml.use_c_lib_yaml = use_c_lib_yaml
            reference = self.EXAMPLE_ENTRY_DICT.copy()
            entries = YAMLParser().parse(self.EXAMPLE_YAML_FILE)
            entry = list(entries.values())[0]
            assert entry.data == reference
        finally:
            config.defaults()

    def test_raise_missing_file(self) -> None:
        """Test assertion is raised for missing file."""
        with pytest.raises(FileNotFoundError):
            YAMLParser().parse("test/missing_file.yaml")

    def test_event_pre_yaml_parse(self) -> None:
        """Tests the PreYAMLParse event."""

        @Event.PreYAMLParse.subscribe
        def hook(string: str) -> Optional[str]:
            return self.EXAMPLE_YAML_FILE

        assert Event.PreYAMLParse.validate()

        reference = self.EXAMPLE_ENTRY_DICT.copy()
        entries = YAMLParser().parse("Hello world!")
        entry = list(entries.values())[0]
        assert entry.data == reference

    def test_event_post_yaml_parse(self) -> None:
        """Tests the PostYAMLParse event."""

        @Event.PostYAMLParse.subscribe
        def hook(bib: Dict[str, Entry]) -> None:
            bib["Cao_2019"].data["month"] = "August"

        reference = self.EXAMPLE_ENTRY_DICT.copy()
        reference["month"] = "August"

        assert Event.PostYAMLParse.validate()

        entries = YAMLParser().parse(self.EXAMPLE_YAML_FILE)
        entry = list(entries.values())[0]
        assert entry.data == reference

    def test_event_pre_yaml_dump(self) -> None:
        """Tests the PreYAMLDump event."""

        @Event.PreYAMLDump.subscribe
        def hook(entry: Entry) -> None:
            entry.label = "Cao2019"

        assert Event.PreYAMLDump.validate()

        entry = Entry("Cao_2019", self.EXAMPLE_ENTRY_DICT.copy())
        entry_str = YAMLParser().dump(entry)
        assert cast(str, entry_str).split("\n")[1] == "Cao2019:"

    def test_event_post_yaml_dump(self) -> None:
        """Tests the PostYAMLDump event."""

        @Event.PostYAMLDump.subscribe
        def hook(string: str) -> Optional[str]:
            return "Hello world!"

        assert Event.PostYAMLDump.validate()

        entry = Entry("Cao_2019", self.EXAMPLE_ENTRY_DICT.copy())
        entry_str = YAMLParser().dump(entry)
        assert entry_str == "Hello world!"



"""coBib's file downloader utility."""

from __future__ import annotations

import asyncio
import logging
import re
import tempfile
from pathlib import Path
from typing import Dict, Optional, Type

import requests
from rich.console import Console
from rich.progress import DownloadColumn, Progress, SpinnerColumn, TimeElapsedColumn
from textual.app import App

from cobib.config import Event, config

from .rel_path import RelPath

LOGGER = logging.getLogger(__name__)


class FileDownloader:
    """The file downloader singleton.

    This utility centralizes the downloading of associated files.
    """

    _instance: Optional[FileDownloader] = None
    """The singleton instance of this class."""

    def __new__(cls) -> FileDownloader:
        """Singleton constructor.

        This method gets called when accessing `FileDownloader` and enforces the singleton pattern
        implemented by this class.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    progress: Type[Progress] = Progress
    """The type of progress bar to use when displaying the downloading progress."""

    console: Console | App[None] = Console()
    """The object via which to print output."""

    _PDF_MARKER = bytes("%PDF", "utf-8")
    """A marker which the downloaded file's beginning is checked against, to determine that it is
    indeed a PDF file."""

    @staticmethod
    def _assert_pdf(content: bytes) -> bool:
        """Asserts that the `content` starts with the `_PDF_MARKER`.

        Args:
            content: the string of bytes to check.

        Returns:
            Whether the `content` matches.
        """
        if not content.startswith(FileDownloader._PDF_MARKER):
            LOGGER.warning("The URL did not provide a PDF file. Aborting download!")
            return False
        return True

    @staticmethod
    def _unlink(path: RelPath) -> None:
        """Remove a file and ignore any error.

        Args:
            path: the file to remove.
        """
        path.path.unlink(missing_ok=True)

    @staticmethod
    async def download(
        url: str,
        label: str,
        folder: Optional[str] = None,
        overwrite: bool = False,
        headers: Optional[Dict[str, str]] = None,
    ) -> Optional[RelPath]:
        # pylint: disable=too-many-function-args,unexpected-keyword-arg,no-member
        """Downloads a file.

        The path of the downloaded file is `folder/label.pdf`. The path can be configured via
        `cobib.config.config.FileDownloaderConfig.default_location`.

        Args:
            url: the link to the file to be downloaded.
            label: the name of the entry.
            folder: an optional folder where the downloaded file will be stored.
            overwrite: whether or not to overwrite an existing file.
            headers: optional headers for the download `GET` request.

        Returns:
            The `RelPath` to the downloaded file. If downloading was not successful, `None` is
            returned.
        """
        if folder is None:
            folder = config.utils.file_downloader.default_location

        hook_result = Event.PreFileDownload.fire(url, label, folder, headers)
        if hook_result is not None:
            url, label, folder, headers = hook_result

        path = RelPath(Path(f"{folder}/{label}").with_suffix(".pdf"))

        backup = None
        if path.path.exists():
            if not overwrite:
                LOGGER.warning(
                    "A file at '%s' already exists! Using that rather than downloading.", path
                )
                return path
            backup = FileDownloader._backup_file(path)

        url = FileDownloader._map_url(url)

        with open(path.path, "wb") as file:
            LOGGER.info("Downloading %s to %s", url, path)

            try:
                response = requests.get(url, timeout=10, stream=True, headers=headers)
                total_length = response.headers.get("content-length", None)
                total_length = int(total_length) if total_length is not None else None
            except requests.exceptions.RequestException as err:
                msg = f"An Exception occurred while downloading the file located at {url}"
                LOGGER.warning(msg)
                LOGGER.error(err)
                FileDownloader._recover(path, backup)
                return None

            progress_bar = FileDownloader.progress(
                SpinnerColumn(),
                *Progress.get_default_columns(),
                TimeElapsedColumn(),
                DownloadColumn(),
            )
            progress_bar.start()

            if isinstance(FileDownloader.console, App):
                # pylint: disable=assignment-from-no-return,unpacking-non-sequence
                _, await_mount = FileDownloader.console.print(  # type: ignore[attr-defined]
                    progress_bar
                )
                await await_mount

            task = progress_bar.add_task("Downloading...", total=total_length)

            accumulated_length = 0

            if total_length is None:
                if not FileDownloader._assert_pdf(response.content):
                    FileDownloader._recover(path, backup)
                    progress_bar.stop()
                    return None
                file.write(response.content)
            else:
                for data in response.iter_content(chunk_size=4096):
                    if accumulated_length == 0 and not FileDownloader._assert_pdf(data):
                        FileDownloader._recover(path, backup)
                        progress_bar.stop()
                        return None
                    accumulated_length += len(data)
                    progress_bar.advance(task, len(data))
                    await asyncio.sleep(0)
                    file.write(data)

            progress_bar.stop()
            if isinstance(FileDownloader.console, App):
                progress_bar.set_timer(5.0, progress_bar.remove)  # type: ignore[attr-defined]

            msg = f"Successfully downloaded {path}"
            print(msg)
            LOGGER.info(msg)

            path = Event.PostFileDownload.fire(path) or path

            return path

    @staticmethod
    def _map_url(url: str) -> str:
        """Maps a URL according to `cobib.config.config.FileDownloaderConfig.url_map`.

        Args:
            url: the URL to be mapped.

        Returns:
            The mapped URL.
        """
        for pattern_url, repl_url in config.utils.file_downloader.url_map.items():
            if re.match(pattern_url, url):
                new_url: str = re.sub(pattern_url, repl_url, url)
                LOGGER.info(
                    "Matched the file's URL to your pattern URL %s and replaced it to become %s",
                    pattern_url,
                    new_url,
                )
                return new_url
        return url

    @staticmethod
    # pylint: disable=unsubscriptable-object
    def _backup_file(path: RelPath) -> tempfile._TemporaryFileWrapper[bytes]:
        """Create a backup of an existing file.

        Args:
            path: the path to the file to be backed up.

        Returns:
            The temporary backup file.
        """
        # we make a copy of the existing file in case downloading a new one fails
        backup = tempfile.NamedTemporaryFile(delete=False)  # pylint: disable=consider-using-with
        backup.write(path.path.read_bytes())
        backup.seek(0)
        return backup

    @staticmethod
    # pylint: disable=unsubscriptable-object
    def _recover(path: RelPath, backup: Optional[tempfile._TemporaryFileWrapper[bytes]]) -> None:
        """Recovers from a backup file.

        If not `backup` exists, the file location which was supposed to be recovered is properly
        removed.

        Args:
            path: the path to the file to be recovered.
            backup: the temporary backup file.
        """
        FileDownloader._unlink(path)
        if backup is not None:
            path.path.write_bytes(backup.read())
            backup.close()
            Path(backup.name).unlink()



"""coBib's command-line interface.

This class provides the main entry point to coBib. It exposes all the commands implemented in
`cobib.commands` to the end-user and leverages [`rich`](https://github.com/textualize/rich) to
produce beautiful output.
"""

import argparse
import asyncio
import inspect
import logging
import sys
from inspect import iscoroutinefunction
from typing import Any

from rich.console import Console
from typing_extensions import override

from cobib import __version__, commands
from cobib.config import config
from cobib.database import Database
from cobib.ui.tui import TUI
from cobib.ui.ui import UI
from cobib.utils.logging import print_changelog

LOGGER = logging.getLogger(__name__)


class CLI(UI):
    """The CLI class.

    In addition to the global arguments documented by the base class, the following are supported:

      * `--version`: prints the coBib version and quits.
      * `command`: a single positional argument indicating the name of the command to run. When this
        is omitted, the `cobib.ui.tui.TUI` gets started.
    """

    @override
    def add_extra_parser_arguments(self) -> None:
        self.parser.add_argument("--version", action="version", version=f"%(prog)s v{__version__}")

        subcommands = [cls.name for _, cls in inspect.getmembers(commands) if inspect.isclass(cls)]
        self.parser.add_argument(
            "command", help="subcommand to be called", choices=subcommands, nargs="?"
        )
        self.parser.add_argument("args", nargs=argparse.REMAINDER)

    @override
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.init_argument_parser(
            prog="coBib",
            description=(
                "Cobib input arguments.\nIf no arguments are given, the TUI will start as a "
                "default."
            ),
        )

    # pylint: disable=invalid-overridden-method
    @override
    async def run(self) -> None:  # type: ignore[override]
        arguments = self.parse_args()

        console = Console()

        if not arguments.porcelain:
            # print latest changelog
            changelog = print_changelog(__version__, config.logging.version)
            if changelog is not None:
                console.print(changelog)

        if arguments.command == "init":
            # the database file may not exist yet, thus we ensure to execute the command before
            # trying to read the database file
            subcmd = getattr(commands, "InitCommand")(*arguments.args)
            subcmd.execute()
            return

        # initialize database
        Database()

        if not arguments.command:
            task = asyncio.create_task(TUI().run_async())  # type: ignore[abstract]
            await task
            sys.exit()
        else:
            subcmd = getattr(commands, arguments.command.title() + "Command")(*arguments.args)
            if iscoroutinefunction(subcmd.execute):
                await subcmd.execute()
            else:
                subcmd.execute()
            if arguments.porcelain:
                output = subcmd.render_porcelain()
                for line in output:
                    print(line)
            else:
                renderable = subcmd.render_rich()
                if renderable is not None:
                    console.print(renderable)
